# Copyright (C) 2005-2006 Dimitur Kirov <dkirov AT gmail.com>
#                         Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2005-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2006-2008 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2008 Jonathan Schleifer <js-gajim AT webkeks.org>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

import hashlib
import logging
import os
import socket
import struct
import sys
import time
from errno import EAFNOSUPPORT
from errno import EINPROGRESS
from errno import EINTR
from errno import EISCONN
from errno import ENOBUFS
from errno import EWOULDBLOCK

from gajim.common import app
from gajim.common.file_props import FilesProp

# from nbxmpp.idlequeue import IdleObject


log = logging.getLogger('gajim.c.socks5')
MAX_BUFF_LEN = 65536
# after foo seconds without activity label transfer as 'stalled'
STALLED_TIMEOUT = 10
# after foo seconds of waiting to connect, disconnect from
# streamhost and try next one
CONNECT_TIMEOUT = 30
# nothing received for the last foo seconds - stop transfer
# if it is 0, then transfer will wait forever
READ_TIMEOUT = 180
# nothing sent for the last foo seconds - stop transfer
# if it is 0, then transfer will wait forever
SEND_TIMEOUT = 180


class SocksQueue:
    """
    Queue for all file requests objects
    """

    def __init__(self, idlequeue, complete_transfer_cb=None,
    progress_transfer_cb=None, error_cb=None):
        self.connected = 0
        self.readers = {}
        self.senders = {}
        self.idx = 1
        self.listener = None
        self.sha_handlers = {}
        # handle all io events in the global idle queue, instead of processing
        # each foo seconds
        self.idlequeue = idlequeue
        self.complete_transfer_cb = complete_transfer_cb
        self.progress_transfer_cb = progress_transfer_cb
        self.error_cb = error_cb
        self.on_success = {} # {id: cb}
        self.on_failure = {} # {id: cb}

    def start_listener(self, port, sha_str, sha_handler, file_props,
                       typ='sender'):
        """
        Start waiting for incoming connections on (host, port) and do a socks5
        authentication using sid for generated SHA
        """
        log.debug('Start listening for socks5 connection')
        sid = file_props.sid
        self.sha_handlers[sha_str] = (sha_handler, sid)
        if self.listener is None or not self.listener.connections:
            self.listener = Socks5Listener(self.idlequeue, port, file_props)
            self.listener.queue = self
            self.listener.bind()
        else:
            # There is already a listener, we update the file's information
            # on the new connection.
            self.listener.file_props = file_props
        self.connected += 1
        return self.listener

    def send_success_reply(self, file_props, streamhost):
        if file_props.streamhost_used is True:
            for proxy in file_props.proxyhosts:
                if proxy['host'] == streamhost['host']:
                    self.on_success[file_props.transport_sid](proxy)
                    return 1
            return 0
        for host in file_props.streamhosts:
            if host['state'] == 1:
                # on_success callback already called for another host
                return 0 # return 0 to disconnect this one
        streamhost['state'] = 1
        self.on_success[file_props.transport_sid](streamhost)
        return 1

    def connect_to_hosts(self, account, transport_sid, on_success=None,
                         on_failure=None, receiving=True):
        self.on_success[transport_sid] = on_success
        self.on_failure[transport_sid] = on_failure
        file_props = FilesProp.getFilePropByTransportSid(account, transport_sid)
        file_props.failure_cb = on_failure
        streamhosts_to_test = []
        # Remove local IPs to not connect to ourself
        for streamhost in file_props.streamhosts:
            if streamhost['host'] == '127.0.0.1' or streamhost['host'] == '::1':
                continue
            streamhosts_to_test.append(streamhost)
        if not streamhosts_to_test:
            on_failure(file_props.transport_sid)
        # add streamhosts to the queue
        for streamhost in streamhosts_to_test:
            if receiving:
                if 'candidate_id' in streamhost:
                    log.debug('Trying to connect as receiver to cid %s',
                              streamhost['candidate_id'])
                else:
                    log.debug('Trying to connect as receiver to jid %s',
                              streamhost['jid'])
                file_props.type_ = 'r'
                socks5obj = Socks5ReceiverClient(self.idlequeue, streamhost,
                    transport_sid, file_props)
                self.add_sockobj(account, socks5obj)
            else:
                if 'candidate_id' in streamhost:
                    log.debug('Trying to connect as sender to cid %s',
                              streamhost['candidate_id'])
                else:
                    log.debug('Trying to connect as sender to jid %s',
                              streamhost['jid'])
                if file_props.sha_str:
                    idx = file_props.sha_str
                else:
                    idx = self.idx
                    self.idx = self.idx + 1
                file_props.type_ = 's'
                if 'type' in streamhost and streamhost['type'] == 'proxy':
                    file_props.is_a_proxy = True
                    file_props.proxy_sender = streamhost['target']
                    file_props.proxy_receiver = streamhost['initiator']
                socks5obj = Socks5SenderClient(self.idlequeue, idx,
                    self, _sock=None, host=str(streamhost['host']),
                    port=int(streamhost['port']),
                    connected=False, file_props=file_props,
                    initiator=streamhost['initiator'],
                    target=streamhost['target'])
                socks5obj.streamhost = streamhost
                self.add_sockobj(account, socks5obj)

            streamhost['idx'] = socks5obj.queue_idx

    def _socket_connected(self, streamhost, file_props):
        """
        Called when there is a host connected to one of the senders's
        streamhosts. Stop other attempts for connections
        """
        if 'candidate_id' in streamhost:
            log.debug('Connected to cid %s', streamhost['candidate_id'])
        else:
            log.debug('Connected to jid %s', streamhost['jid'])
        for host in file_props.streamhosts:
            if host != streamhost and 'idx' in host:
                if host['state'] == 1:
                    # remove current
                    if file_props.type_ == 's':
                        self.remove_sender(streamhost['idx'], False)
                    else:
                        self.remove_receiver(streamhost['idx'])
                    return
                # set state -2, meaning that this streamhost is stopped,
                # but it may be connected later
                if host['state'] >= 0:
                    if file_props.type_ == 's':
                        self.remove_sender(host['idx'], False)
                    else:
                        self.remove_receiver(host['idx'])
                    host['idx'] = -1
                    host['state'] = -2

    def reconnect_client(self, client, streamhost):
        """
        Check the state of all streamhosts and if all has failed, then emit
        connection failure cb. If there are some which are still not connected
        try to establish connection to one of them
        """
        self.idlequeue.remove_timeout(client.fd)
        self.idlequeue.unplug_idle(client.fd)
        file_props = client.file_props
        streamhost['state'] = -1
        # boolean, indicates that there are hosts, which are not tested yet
        unused_hosts = False
        for host in file_props.streamhosts:
            if 'idx' in host:
                if host['state'] >= 0:
                    return
                if host['state'] == -2:
                    unused_hosts = True
        if unused_hosts:
            for host in file_props.streamhosts:
                if host['state'] == -2:
                    host['state'] = 0
                    # FIXME: make the sender reconnect also
                    client = Socks5ReceiverClient(self.idlequeue, host,
                        client.sid, file_props)
                    self.add_sockobj(client.account, client)
                    host['idx'] = client.queue_idx
            # we still have chances to connect
            return
        if file_props.received_len == 0:
            # there are no other streamhosts and transfer hasn't started
            self._connection_refused(streamhost, file_props, client.queue_idx)
        else:
            # transfer stopped, it is most likely stopped from sender
            client.disconnect()
            file_props.error = -1
            self.process_result(-1, client)

    def _connection_refused(self, streamhost, file_props, idx):
        """
        Called when we loose connection during transfer
        """
        if 'candidate_id' in streamhost:
            log.debug('Connection refused to cid %s',
                      streamhost['candidate_id'])
        else:
            log.debug('Connection refused to jid %s', streamhost['jid'])
        if file_props is None:
            return
        streamhost['state'] = -1
        # FIXME: should only the receiver be remove? what if we are sending?
        self.remove_receiver(idx, False)
        for host in file_props.streamhosts:
            if host['state'] != -1:
                return
        self.readers = {}
        # failure_cb exists - this means that it has never been called
        if file_props.failure_cb:
            file_props.failure_cb(file_props.transport_sid)
            file_props.failure_cb = None

    def add_sockobj(self, account, sockobj):
        """
        Add new file a sockobj type receiver or sender, and use it to connect
        to server
        """
        if sockobj.file_props.type_ == 'r':
            self._add(sockobj, self.readers, sockobj.file_props, self.idx)
        else:
            self._add(sockobj, self.senders, sockobj.file_props, self.idx)
        sockobj.queue_idx = self.idx
        sockobj.queue = self
        sockobj.account = account
        self.idx += 1
        result = sockobj.connect()
        self.connected += 1
        if result is not None:
            result = sockobj.main()
            self.process_result(result, sockobj)
            return 1
        return None

    def _add(self, sockobj, sockobjects, file_props, hash_):
        '''
        Adds the sockobj to the current list of sockobjects
        '''
        keys = (file_props.transport_sid, file_props.name, hash_)
        sockobjects[keys] = sockobj

    def result_sha(self, sha_str, idx):
        if sha_str in self.sha_handlers:
            props = self.sha_handlers[sha_str]
            props[0](props[1], idx)

    def activate_proxy(self, idx):
        if not self.isHashInSockObjs(self.senders, idx):
            return
        for key in list(self.senders):
            if idx in key:
                sender = self.senders[key]
                if sender.file_props.type_ != 's':
                    return
                sender.state = 6
                if sender.connected:
                    sender.file_props.error = 0
                    sender.file_props.disconnect_cb = sender.disconnect
                    sender.file_props.started = True
                    sender.file_props.completed = False
                    sender.file_props.paused = False
                    sender.file_props.stalled = False
                    sender.file_props.elapsed_time = 0
                    sender.file_props.last_time = time.time()
                    sender.file_props.received_len = 0
                    sender.pauses = 0
                    # start sending file to proxy
                    self.idlequeue.set_read_timeout(sender.fd, STALLED_TIMEOUT)
                    self.idlequeue.plug_idle(sender, True, False)
                    result = sender.write_next()
                    self.process_result(result, sender)

    def send_file(self, file_props, account, mode):
        for key in list(self.senders.keys()):
            if file_props.name in key and file_props.transport_sid in key \
            and self.senders[key].mode == mode:
                log.info('socks5: sending file')
                sender = self.senders[key]
                file_props.streamhost_used = True
                sender.account = account
                sender.file_props = file_props
                result = sender.send_file()
                self.process_result(result, sender)

    def isHashInSockObjs(self, sockobjs, hash_):
        '''
        It tells whether there is a particular hash in sockobjs or not
        '''
        for key in sockobjs:
            if hash_ in key:
                return True
        return False

    def on_connection_accepted(self, sock, listener):
        sock_hash = hash(sock)
        if listener.file_props.type_ == 's' and \
        not self.isHashInSockObjs(self.senders, sock_hash):
            sockobj = Socks5SenderServer(self.idlequeue, sock_hash, self,
                sock[0], sock[1][0], sock[1][1],
                file_props=listener.file_props)
            self._add(sockobj, self.senders, listener.file_props, sock_hash)
            # Start waiting for data
            self.idlequeue.plug_idle(sockobj, False, True)
            self.connected += 1
        if listener.file_props.type_ == 'r' and \
        not self.isHashInSockObjs(self.readers, sock_hash):
            sh = {}
            sh['host'] = sock[1][0]
            sh['port'] = sock[1][1]
            sh['initiator'] = None
            sh['target'] = None
            sockobj = Socks5ReceiverServer(idlequeue=self.idlequeue,
                streamhost=sh, transport_sid=None,
                file_props=listener.file_props)

            self._add(sockobj, self.readers, listener.file_props, sock_hash)
            sockobj.set_sock(sock[0])
            sockobj.queue = self
            self.connected += 1

    def process_result(self, result, actor):
        """
        Take appropriate actions upon the result:
                [ 0, - 1 ] complete/end transfer
                [ > 0 ] send progress message
                [ None ] do nothing
        """
        if result is None:
            return
        if result in (0, -1) and self.complete_transfer_cb is not None:
            account = actor.account
            if account is None and actor.file_props.tt_account:
                account = actor.file_props.tt_account
            self.complete_transfer_cb(account, actor.file_props)
        elif self.progress_transfer_cb is not None:
            self.progress_transfer_cb(actor.account, actor.file_props)

    def remove_receiver_by_key(self, key, do_disconnect=True):
        reader = self.readers[key]
        self.idlequeue.unplug_idle(reader.fd)
        self.idlequeue.remove_timeout(reader.fd)
        if do_disconnect:
            reader.disconnect()
        else:
            if reader.streamhost is not None:
                reader.streamhost['state'] = -1
        del self.readers[key]

    def remove_sender_by_key(self, key, do_disconnect=True):
        sender = self.senders[key]
        if do_disconnect:
            sender.disconnect()
        self.idlequeue.unplug_idle(sender.fd)
        self.idlequeue.remove_timeout(sender.fd)
        del self.senders[key]
        if self.connected > 0:
            self.connected -= 1

    def remove_receiver(self, idx, do_disconnect=True, remove_all=False):
        """
        Remove receiver from the list and decrease the number of active
        connections with 1
        """
        if idx != -1:
            for key in list(self.readers.keys()):
                if idx in key:
                    self.remove_receiver_by_key(
                        key, do_disconnect=do_disconnect)
                    if not remove_all:
                        break

    def remove_sender(self, idx, do_disconnect=True, remove_all=False):
        """
        Remove sender from the list of senders and decrease the number of active
        connections with 1
        """
        if idx != -1:
            for key in list(self.senders.keys()):
                if idx in key:
                    self.remove_sender_by_key(key, do_disconnect=do_disconnect)
                    if not remove_all:
                        break
            if not self.senders and self.listener is not None:
                self.listener.disconnect()
                self.listener = None
                self.connected -= 1

    def remove_by_mode(self, transport_sid, mode, do_disconnect=True):
        for (key, sock) in self.senders.copy().items():
            if key[0] == transport_sid and sock.mode == mode:
                self.remove_sender_by_key(key)
        for (key, sock) in self.readers.copy().items():
            if key[0] == transport_sid and sock.mode == mode:
                self.remove_receiver_by_key(key)

    def remove_server(self, transport_sid, do_disconnect=True):
        self.remove_by_mode(transport_sid, 'server')

    def remove_client(self, transport_sid, do_disconnect=True):
        self.remove_by_mode(transport_sid, 'client')

    def remove_other_servers(self, host_to_keep):
        for (key, sock) in self.senders.copy().items():
            if sock.host != host_to_keep and sock.mode == 'server':
                self.remove_sender_by_key(key)

class Socks5:
    def __init__(self, idlequeue, host, port, initiator, target, sid):
        if host is not None:
            try:
                self.host = host
                self.ais = socket.getaddrinfo(host, port, socket.AF_UNSPEC,
                    socket.SOCK_STREAM)
            except socket.gaierror:
                self.ais = None
        self.idlequeue = idlequeue
        self.fd = -1
        self.port = port
        self.initiator = initiator
        self.target = target
        self.sid = sid
        self._sock = None
        self.account = None
        self.state = 0 # not connected
        self.pauses = 0
        self.size = 0
        self.remaining_buff = b''
        self.file = None
        self.connected = False
        self.mode = ''

    def _is_connected(self):
        if self.state < 5:
            return False
        return True

    def connect(self):
        """
        Create the socket and plug it to the idlequeue
        """
        if self.ais is None:
            return None
        for ai in self.ais:
            try:
                self._sock = socket.socket(*ai[:3])
                # this will not block the GUI
                self._sock.setblocking(False)
                self._server = ai[4]
                break
            except socket.error as e:
                if e.errno == EINPROGRESS:
                    break
                # for all other errors, we try other addresses
                continue
        self.fd = self._sock.fileno()
        self.state = 0 # about to be connected
        self.idlequeue.plug_idle(self, True, False)
        self.do_connect()
        self.idlequeue.set_read_timeout(self.fd, CONNECT_TIMEOUT)
        return None

    def do_connect(self):
        try:
            self._sock.connect(self._server)
            self._send = self._sock.send
            self._recv = self._sock.recv
        except Exception as ee:
            errnum = ee.errno
            self.connect_timeout += 1
            if errnum == 111 or self.connect_timeout > 1000:
                self.queue._connection_refused(self.streamhost, self.file_props,
                    self.queue_idx)
                self.connected = False
                return None
            # win32 needs this
            if errnum not in  (10056, EISCONN) or self.state != 0:
                return None
            # socket is already connected
            self._sock.setblocking(False)
            self._send = self._sock.send
            self._recv = self._sock.recv
        self.buff = ''
        self.connected = True
        self.file_props.connected = True
        self.file_props.disconnect_cb = self.disconnect
        self.file_props.paused = False
        self.state = 1 # connected
        # stop all others connections to sender's streamhosts
        self.queue._socket_connected(self.streamhost, self.file_props)
        self.idlequeue.plug_idle(self, True, False)
        return 1 # we are connected

    def read_timeout(self):
        self.idlequeue.remove_timeout(self.fd)
        if self.state > 5:
            # no activity for foo seconds
            if self.file_props.stalled is False:
                self.file_props.stalled = True
                self.queue.process_result(-1, self)
                if not self.file_props.received_len:
                    self.file_props.received_len = 0
                if SEND_TIMEOUT > 0:
                    self.idlequeue.set_read_timeout(self.fd, SEND_TIMEOUT)
            else:
                # stop transfer, there is no error code for this
                self.pollend()
        else:
            if self.mode == 'client':
                self.queue.reconnect_client(self, self.streamhost)

    def open_file_for_reading(self):
        if self.file is None:
            try:
                self.file = open(self.file_props.file_name, 'rb')
                if self.file_props.offset:
                    self.size = self.file_props.offset
                    self.file.seek(self.size)
                    self.file_props.received_len = self.size
            except IOError as e:
                self.close_file()
                raise IOError(str(e))

    def close_file(self):
        # Close file we're sending from
        if self.file:
            if not self.file.closed:
                try:
                    self.file.close()
                except Exception:
                    pass
            self.file = None
        # Close file we're receiving into
        if self.file_props.fd and self.state >= 7:
            try:
                self.file_props.fd.close()
            except Exception:
                pass

    def get_fd(self):
        """
        Test if file is already open and return its fd, or just open the file
        and return the fd
        """
        if self.file_props.fd:
            fd = self.file_props.fd
        else:
            offset = 0
            opt = 'wb'
            if self.file_props.offset:
                offset = self.file_props.offset
                opt = 'ab'
            fd = open(self.file_props.file_name, opt)
            self.file_props.fd = fd
            self.file_props.elapsed_time = 0
            self.file_props.last_time = time.time()
            self.file_props.received_len = offset
        return fd

    def rem_fd(self, fd):
        if self.file_props.fd:
            self.file_props.fd = None
        try:
            fd.close()
        except Exception:
            pass

    def receive(self):
        """
        Read small chunks of data. Call owner's disconnected() method if
        appropriate
        """
        received = b''
        try:
            add = self._recv(64)
        except Exception:
            add = b''
        received += add
        if not add:
            self.disconnect()
        return add

    def send_raw(self, raw_data):
        """
        Write raw outgoing data
        """
        try:
            self._send(raw_data)
        except Exception:
            self.disconnect()
        return len(raw_data)

    def write_next(self):
        if self.remaining_buff != b'':
            buff = self.remaining_buff
        else:
            try:
                self.open_file_for_reading()
            except IOError:
                self.state = 8 # end connection
                self.disconnect()
                self.file_props.error = -7 # unable to read from file
                return -1
            buff = self.file.read(MAX_BUFF_LEN)
        if buff:
            lenn = 0
            try:
                lenn = self._send(buff)
            except socket.error as err:
                if err.errno not in (EINTR, ENOBUFS, EWOULDBLOCK):
                    return self._on_send_exception()
            except Exception as err:
                log.error(err)
                return self._on_send_exception()

            self.size += lenn
            current_time = time.time()
            self.file_props.elapsed_time += current_time - \
                self.file_props.last_time
            self.file_props.last_time = current_time
            self.file_props.received_len = self.size
            if self.size >= self.file_props.size:
                self.state = 8 # end connection
                self.file_props.error = 0
                self.disconnect()
                return -1
            if lenn != len(buff):
                self.remaining_buff = buff[lenn:]
            else:
                self.remaining_buff = b''
            self.state = 7 # continue to write in the socket
            if lenn == 0:
                return None
            self.file_props.stalled = False
            return lenn

        self.state = 8 # end connection
        self.disconnect()
        return -1

    def _on_send_exception(self):
        # peer stopped reading
        self.state = 8 # end connection
        self.disconnect()
        self.file_props.error = -1
        return -1

    def get_file_contents(self, timeout):
        """
        Read file contents from socket and write them to file
        """
        if self.file_props is None or not self.file_props.file_name:
            self.file_props.error = -2
            return None
        fd = None
        if self.remaining_buff != b'':
            try:
                fd = self.get_fd()
            except IOError:
                self.disconnect()
                self.file_props.error = -6 # file system error
                return 0
            fd.write(self.remaining_buff)
            lenn = len(self.remaining_buff)
            current_time = time.time()
            self.file_props.elapsed_time += current_time - \
                self.file_props.last_time
            self.file_props.last_time = current_time
            self.file_props.received_len += lenn
            self.remaining_buff = b''
            if self.file_props.received_len == self.file_props.size:
                self.rem_fd(fd)
                self.disconnect()
                self.file_props.error = 0
                self.file_props.completed = True
                return 0
        else:
            try:
                fd = self.get_fd()
            except IOError:
                self.disconnect()
                self.file_props.error = -6 # file system error
                return 0
            try:
                buff = self._recv(MAX_BUFF_LEN)
            except Exception:
                buff = b''
            current_time = time.time()
            self.file_props.elapsed_time += current_time - \
                self.file_props.last_time
            self.file_props.last_time = current_time
            self.file_props.received_len += len(buff)
            if not buff:
                # Transfer stopped  somehow:
                # reset, paused or network error
                self.rem_fd(fd)
                self.disconnect()
                self.file_props.error = -1
                return 0
            try:
                fd.write(buff)
            except IOError:
                self.rem_fd(fd)
                self.disconnect()
                self.file_props.error = -6 # file system error
                return 0
            if self.file_props.received_len >= self.file_props.size:
                # transfer completed
                self.rem_fd(fd)
                self.disconnect()
                self.file_props.error = 0
                self.file_props.completed = True
                return 0
            # return number of read bytes. It can be used in progressbar
        if fd is not None:
            self.file_props.stalled = False
        if fd is None and self.file_props.stalled is False:
            return None
        if self.file_props.received_len:
            if self.file_props.received_len != 0:
                return self.file_props.received_len
        return None

    def disconnect(self, *args, **kwargs):
        """
        Close open descriptors and remover socket descr. from idleque
        """
        # be sure that we don't leave open file
        self.close_file()
        self.idlequeue.remove_timeout(self.fd)
        self.idlequeue.unplug_idle(self.fd)
        if self.mode == 'server' and self.queue.listener:
            try:
                self.queue.listener.connections.remove(self._sock)
            except ValueError:
                pass # Not in list
            if self.queue.listener.connections == []:
                self.queue.listener.disconnect()
        try:
            self._sock.shutdown(socket.SHUT_RDWR)
            self._sock.close()
        except Exception:
            # socket is already closed
            pass
        self.connected = False
        self.fd = -1
        self.state = -1

    def _get_auth_buff(self):
        """
        Message, that we support 1 one auth mechanism: the 'no auth' mechanism
        """
        return struct.pack('!BBB', 0x05, 0x01, 0x00)

    def _parse_auth_buff(self, buff):
        """
        Parse the initial message and create a list of auth mechanisms
        """
        if buff[0] != 5:
            return None
        num_auth = buff[1]
        return list(buff[2:2+num_auth])

    def _get_auth_response(self):
        """
        Socks version(5), number of extra auth methods (we send 0x00 - no auth)
        """
        return struct.pack('!BB', 0x05, 0x00)

    def _get_connect_buff(self):
        """
        Connect request by domain name
        """
        buff = struct.pack('!BBBBB%dsBB' % len(self.host),
            0x05, 0x01, 0x00, 0x03, len(self.host), self.host.encode('utf-8'),
            self.port >> 8, self.port & 0xff)
        return buff

    def _get_request_buff(self, msg, command=0x01):
        """
        Connect request by domain name, sid sha, instead of domain name (jep
        0096)
        """
        if isinstance(msg, str):
            msg = msg.encode('utf-8')
        buff = struct.pack('!BBBBB%dsBB' % len(msg), 0x05, command, 0x00, 0x03,
            len(msg), msg, 0, 0)
        return buff

    def _parse_request_buff(self, buff):
        try: # don't trust on what comes from the outside
            req_type, host_type, = struct.unpack('!xBxB', buff[:4])
            if host_type == 0x01:
                host_arr = struct.unpack('!iiii', buff[4:8])
                host, = '.'.join(str(s) for s in host_arr)
                host_len = len(host)
            elif host_type == 0x03:
                host_len = buff[4]
                host, = struct.unpack('!%ds' % host_len, buff[5:5 + host_len])
            portlen = len(buff[host_len + 5:])
            if portlen == 1:
                port, = struct.unpack('!B', buff[host_len + 5])
            elif portlen == 2:
                port, = struct.unpack('!H', buff[host_len + 5:])
            # file data, comes with auth message (Gaim bug)
            else:
                port, = struct.unpack('!H', buff[host_len + 5: host_len + 7])
                self.remaining_buff = buff[host_len + 7:]
        except Exception:
            return (None, None, None)
        return (req_type, host, port)

    def read_connect(self):
        """
        Connect response: version, auth method
        """
        buff = self._recv().decode('utf-8')
        try:
            version, method = struct.unpack('!BB', buff)
        except Exception:
            version, method = None, None
        if version != 0x05 or method == 0xff:
            self.disconnect()

    def continue_paused_transfer(self):
        if self.state < 5:
            return
        if self.file_props.type_ == 'r':
            self.idlequeue.plug_idle(self, False, True)
        else:
            self.idlequeue.plug_idle(self, True, False)

    def _get_sha1_auth(self):
        """
        Get sha of sid + Initiator jid + Target jid
        """
        if self.file_props.is_a_proxy:
            return hashlib.sha1(('%s%s%s' % (self.sid,
                self.file_props.proxy_sender, self.file_props.proxy_receiver)).\
                encode('utf-8')).hexdigest()
        return hashlib.sha1(('%s%s%s' % (self.sid, self.initiator,
            self.target)).encode('utf-8')).hexdigest()


class Socks5Sender(IdleObject):
    """
    Class for sending file to socket over socks5
    """
    def __init__(self, idlequeue, sock_hash, parent, _sock, host=None,
                 port=None, connected=True, file_props=None):
        IdleObject.__init__(self)
        self.queue_idx = sock_hash
        self.queue = parent
        self.file_props = file_props
        self.proxy = False
        self._sock = _sock
        if _sock is not None:
            self._sock.setblocking(False)
            self.fd = _sock.fileno()
            self._recv = _sock.recv
            self._send = _sock.send
        self.connected = connected
        self.state = 1 # waiting for first bytes
        self.connect_timeout = 0

        self.file_props.error = 0
        self.file_props.disconnect_cb = self.disconnect
        self.file_props.started = True
        self.file_props.completed = False
        self.file_props.paused = False
        self.file_props.continue_cb = self.continue_paused_transfer
        self.file_props.stalled = False
        self.file_props.connected = True
        self.file_props.elapsed_time = 0
        self.file_props.last_time = time.time()
        self.file_props.received_len = 0

    def start_transfer(self):
        """
        Send the file
        """
        return self.write_next()

    def set_connection_sock(self, _sock):
        self._sock = _sock
        self._sock.setblocking(False)
        self.fd = _sock.fileno()
        self._recv = _sock.recv
        self._send = _sock.send
        self.connected = True
        self.state = 1 # waiting for first bytes
        self.file_props = None
        # start waiting for data
        self.idlequeue.plug_idle(self, False, True)

    def send_file(self):
        """
        Start sending the file over verified connection
        """
        self.pauses = 0
        self.state = 7
        # plug for writing
        self.idlequeue.plug_idle(self, True, False)
        return self.write_next() # initial for nl byte

    def disconnect(self, cb=True):
        """
        Close the socket
        """
        # close connection and remove us from the queue
        Socks5.disconnect(self)
        if self.file_props is not None:
            self.file_props.connected = False
            self.file_props.disconnect_cb = None
        if self.queue is not None:
            self.queue.remove_sender(self.queue_idx, False)


class Socks5Receiver(IdleObject):
    def __init__(self, idlequeue, streamhost, sid, file_props=None):
        IdleObject.__init__(self)
        self.queue_idx = -1
        self.streamhost = streamhost
        self.queue = None
        self.connect_timeout = 0
        self.connected = False
        self.pauses = 0
        self.sid = sid
        self.file_props = file_props
        self.file_props.disconnect_cb = self.disconnect
        self.file_props.error = 0
        self.file_props.started = True
        self.file_props.completed = False
        self.file_props.paused = False
        self.file_props.continue_cb = self.continue_paused_transfer
        self.file_props.stalled = False
        self.file_props.received_len = 0

    def receive_file(self):
        """
        Start receiving the file over verified connection
        """
        if self.file_props.started:
            return
        self.file_props.error = 0
        self.file_props.disconnect_cb = self.disconnect
        self.file_props.started = True
        self.file_props.completed = False
        self.file_props.paused = False
        self.file_props.continue_cb = self.continue_paused_transfer
        self.file_props.stalled = False
        self.file_props.connected = True
        self.file_props.elapsed_time = 0
        self.file_props.last_time = time.time()
        self.file_props.received_len = 0
        self.pauses = 0
        self.state = 7
        # plug for reading
        self.idlequeue.plug_idle(self, False, True)
        return self.get_file_contents(0) # initial for nl byte

    def start_transfer(self):
        """
        Receive the file
        """
        return self.get_file_contents(0)

    def set_sock(self, _sock):
        self._sock = _sock
        self._sock.setblocking(False)
        self.fd = _sock.fileno()
        self._recv = _sock.recv
        self._send = _sock.send
        self.connected = True
        self.state = 1 # waiting for first bytes
        # start waiting for data
        self.idlequeue.plug_idle(self, False, True)

    def disconnect(self, cb=True):
        """
        Close the socket. Remove self from queue if cb is True
        """
        # close connection
        Socks5.disconnect(self)
        if cb is True:
            self.file_props.disconnect_cb = None
        if self.queue is not None:
            self.queue.remove_receiver(self.queue_idx, False)

class Socks5Server(Socks5):

    def __init__(self, idlequeue, host, port, initiator, target, sid):
        Socks5.__init__(self, idlequeue, host, port, initiator, target, sid)
        self.mode = 'server'

    def main(self):
        """
        Initial requests for verifying the connection
        """
        if self.state == 1: # initial read
            buff = self.receive()
            if not self.connected:
                return -1
            mechs = self._parse_auth_buff(buff)
            if mechs is None:
                return -1 # invalid auth methods received
        elif self.state == 3: # get next request
            buff = self.receive()
            req_type, self.sha_msg = self._parse_request_buff(buff)[:2]
            if req_type != 0x01:
                return -1 # request is not of type 'connect'
        self.state += 1 # go to the next step
        # unplug & plug for writing
        self.idlequeue.plug_idle(self, True, False)
        return None

    def pollin(self):
        self.idlequeue.remove_timeout(self.fd)
        if self.connected:
            if self.state < 5:
                result = self.main()
                if self.state == 4:
                    self.queue.result_sha(self.sha_msg, self.queue_idx)
                if result == -1:
                    self.disconnect()
            elif self.state == 5:
                self.state = 7
                if self.file_props.type_ == 's':
                    # We wait for the end of the negotiation to
                    # send the file
                    self.idlequeue.plug_idle(self, False, False)
                else:
                    # We plug for reading
                    self.idlequeue.plug_idle(self, False, True)
                    return
            elif self.state == 7:
                if self.file_props.paused:
                    self.file_props.continue_cb = \
                        self.continue_paused_transfer
                    self.idlequeue.plug_idle(self, False, False)
                    return
                self.idlequeue.set_read_timeout(self.fd, STALLED_TIMEOUT)
                result = self.start_transfer() # send
                self.queue.process_result(result, self)
        else:
            self.disconnect()

    def pollend(self):
        self.state = 8 # end connection
        self.disconnect()
        self.file_props.error = -1
        self.queue.process_result(-1, self)

    def pollout(self):
        if not self.connected:
            self.disconnect()
            return
        self.idlequeue.remove_timeout(self.fd)
        if self.state == 2: # send reply with desired auth type
            self.send_raw(self._get_auth_response())
        elif self.state == 4: # send positive response to the 'connect'
            self.send_raw(self._get_request_buff(self.sha_msg, 0x00))
        elif self.state == 7:
            if self.file_props.paused:
                self.file_props.continue_cb = self.continue_paused_transfer
                self.idlequeue.plug_idle(self, False, False)
                return
            result = self.start_transfer() # send
            self.queue.process_result(result, self)
            if result is None or result <= 0:
                self.disconnect()
                return
            self.idlequeue.set_read_timeout(self.fd, STALLED_TIMEOUT)
        elif self.state == 8:
            self.disconnect()
            return
        else:
            self.disconnect()
        if self.state < 5:
            self.state += 1
            # unplug and plug this time for reading
            self.idlequeue.plug_idle(self, False, True)


class Socks5Client(Socks5):

    def __init__(self, idlequeue, host, port, initiator, target, sid):
        Socks5.__init__(self, idlequeue, host, port, initiator, target, sid)
        self.mode = 'client'

    def main(self, timeout=0):
        """
        Begin negotiation. on success 'address' != 0
        """
        result = 1
        buff = self.receive()
        if buff == b'':
            # end connection
            self.pollend()
            return
        if self.state == 2: # read auth response
            if buff is None or len(buff) != 2:
                return None
            version, method = struct.unpack('!BB', buff[:2])
            if version != 0x05 or method == 0xff:
                self.disconnect()
        elif self.state == 4: # get approve of our request
            if buff is None:
                return None
            sub_buff = buff[:4]
            if len(sub_buff) < 4:
                return None
            version, address_type = struct.unpack('!BxxB', buff[:4])
            addrlen = 0
            if address_type == 0x03:
                addrlen = buff[4]
                # address = struct.unpack('!%ds' % addrlen, buff[5:addrlen + 5])
                portlen = len(buff[addrlen + 5:])
                # if portlen == 1:
                #     port, = struct.unpack('!B', buff[addrlen + 5])
                # elif portlen == 2:
                #     port, = struct.unpack('!H', buff[addrlen + 5:])
                # else: # Gaim bug :)
                #     port, = struct.unpack('!H', buff[addrlen + 5:addrlen + 7])
                if portlen not in (1, 2):
                    self.remaining_buff = buff[addrlen + 7:]

            self.state = 5 # for senders: init file_props and send '\n'
            if self.queue.on_success:
                result = self.queue.send_success_reply(self.file_props,
                    self.streamhost)
                if self.file_props.type_ == 's' and self.proxy:
                    self.queue.process_result(self.send_file(), self)
                    return
                if result == 0:
                    self.state = 8
                    self.disconnect()
        # for senders: init file_props
        if result == 1 and self.state == 5:
            if self.file_props.type_ == 's':
                self.file_props.error = 0
                self.file_props.disconnect_cb = self.disconnect
                self.file_props.started = True
                self.file_props.completed = False
                self.file_props.paused = False
                self.file_props.stalled = False
                self.file_props.elapsed_time = 0
                self.file_props.last_time = time.time()
                self.file_props.received_len = 0
                self.pauses = 0
                # start sending file contents to socket
                #self.idlequeue.set_read_timeout(self.fd, STALLED_TIMEOUT)
                #self.idlequeue.plug_idle(self, True, False)
                self.idlequeue.plug_idle(self, False, False)
            else:
                # receiving file contents from socket
                self.idlequeue.plug_idle(self, False, True)
            self.file_props.continue_cb = self.continue_paused_transfer
            # we have set up the connection, next - retrieve file
            self.state = 6
        if self.state < 5:
            self.idlequeue.plug_idle(self, True, False)
            self.state += 1
            return None

    def pollin(self):
        self.idlequeue.remove_timeout(self.fd)
        if self.connected:
            if self.file_props.paused:
                self.idlequeue.plug_idle(self, False, False)
                return
            if self.state < 5:
                self.idlequeue.set_read_timeout(self.fd, CONNECT_TIMEOUT)
                self.main(0)
            elif self.state == 5: # wait for proxy reply
                pass
            elif self.file_props.type_ == 'r':
                self.idlequeue.set_read_timeout(self.fd, STALLED_TIMEOUT)
                result = self.start_transfer() # receive
                self.queue.process_result(result, self)
        else:
            self.disconnect()

    def pollout(self):
        self.idlequeue.remove_timeout(self.fd)
        if self.state == 0:
            self.do_connect()
            return
        if self.state == 1: # send initially: version and auth types
            self.send_raw(self._get_auth_buff())
        elif self.state == 3: # send 'connect' request
            self.send_raw(self._get_request_buff(self._get_sha1_auth()))
        elif self.file_props.type_ != 'r':
            if self.file_props.paused:
                self.idlequeue.plug_idle(self, False, False)
                return
            result = self.start_transfer() # send
            self.queue.process_result(result, self)
            return

        self.state += 1
        # unplug and plug for reading
        self.idlequeue.plug_idle(self, False, True)
        self.idlequeue.set_read_timeout(self.fd, CONNECT_TIMEOUT)

    def pollend(self):
        if self.state >= 5:
            # error during transfer
            self.disconnect()
            self.file_props.error = -1
            self.queue.process_result(-1, self)
        else:
            self.queue.reconnect_client(self, self.streamhost)


class Socks5SenderClient(Socks5Client, Socks5Sender):

    def __init__(self, idlequeue, sock_hash, parent, _sock, host=None,
    port=None, connected=True, file_props=None,
    initiator=None, target=None):
        Socks5Client.__init__(self, idlequeue, host, port, initiator, target,
            file_props.transport_sid)
        Socks5Sender.__init__(self, idlequeue, sock_hash, parent, _sock,
            host, port, connected, file_props)


class Socks5SenderServer(Socks5Server, Socks5Sender):

    def __init__(self, idlequeue, sock_hash, parent, _sock, host=None,
    port=None, connected=True, file_props=None):
        Socks5Server.__init__(self, idlequeue, host, port, None, None,
            file_props.transport_sid)
        Socks5Sender.__init__(self, idlequeue, sock_hash, parent, _sock,
            host, port, connected, file_props)


class Socks5ReceiverClient(Socks5Client, Socks5Receiver):
    def __init__(self, idlequeue, streamhost, transport_sid, file_props=None):
        Socks5Client.__init__(self, idlequeue, streamhost['host'],
            int(streamhost['port']), streamhost['initiator'],
            streamhost['target'], transport_sid)
        Socks5Receiver.__init__(self, idlequeue, streamhost, transport_sid,
            file_props)


class Socks5ReceiverServer(Socks5Server, Socks5Receiver):

    def __init__(self, idlequeue, streamhost, transport_sid, file_props=None):
        Socks5Server.__init__(self, idlequeue, streamhost['host'],
            int(streamhost['port']), streamhost['initiator'],
            streamhost['target'], transport_sid)
        Socks5Receiver.__init__(self, idlequeue, streamhost, transport_sid,
            file_props)


class Socks5Listener(IdleObject):
    def __init__(self, idlequeue, port, fp):
        """
        Handle all incoming connections on (0.0.0.0, port)

        This class implements IdleObject, but we will expect
        only pollin events though
        """
        IdleObject.__init__(self)
        self.port = port
        self.ais = socket.getaddrinfo(None, port, socket.AF_UNSPEC,
            socket.SOCK_STREAM, socket.SOL_TCP, socket.AI_PASSIVE)
        self.ais.sort(reverse=True) # Try IPv6 first
        self.queue_idx = -1
        self.idlequeue = idlequeue
        self.queue = None
        self.started = False
        self._sock = None
        self.fd = -1
        self.file_props = fp
        self.connections = []

    def bind(self):
        for ai in self.ais:
            # try the different possibilities (ipv6, ipv4, etc.)
            try:
                self._serv = socket.socket(*ai[:3])
            except socket.error as e:
                if e.errno == EAFNOSUPPORT:
                    self.ai = None
                    continue
                raise
            self._serv.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self._serv.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self._serv.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            # Under windows Vista, we need that to listen on ipv6 AND ipv4
            # Doesn't work under windows XP
            if os.name == 'nt':
                if sys.getwindowsversion().major >= 6:  # Win Vista +
                    # 47 is socket.IPPROTO_IPV6
                    # 27 is socket.IPV6_V6ONLY under windows, but not defined ...
                    self._serv.setsockopt(41, 27, 0)
            # will fail when port as busy, or we don't have rights to bind
            try:
                self._serv.bind(ai[4])
                self.ai = ai
                break
            except Exception:
                self.ai = None
                continue
        if not self.ai:
            log.error('unable to bind to port %s', str(self.port))
            return None
        self._serv.listen(socket.SOMAXCONN)
        self._serv.setblocking(False)
        self.fd = self._serv.fileno()
        self.idlequeue.plug_idle(self, False, True)
        self.started = True

    def pollend(self):
        """
        Called when we stop listening on (host, port)
        """
        self.disconnect()

    def pollin(self):
        """
        Accept a new incoming connection and notify queue
        """
        sock = self.accept_conn()
        self.queue.on_connection_accepted(sock, self)

    def disconnect(self):
        """
        Free all resources, we are not listening anymore
        """
        self.idlequeue.remove_timeout(self.fd)
        self.idlequeue.unplug_idle(self.fd)
        self.fd = -1
        self.state = -1
        self.started = False
        try:
            self._serv.close()
        except Exception:
            pass

    def accept_conn(self):
        """
        Accept a new incoming connection
        """
        _sock = self._serv.accept()
        _sock[0].setblocking(False)
        self.connections.append(_sock[0])
        return _sock



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from datetime import datetime
from datetime import timedelta

from gi.repository import Gtk

from gajim.common.modules.chat_markers import DisplayedMarkerData

from gajim.gtk.conversation.avatar_stack import AvatarStack
from gajim.gtk.conversation.rows.base import BaseRow
from gajim.gtk.widgets import TimeLabel


class DisplayedRow(BaseRow):
    def __init__(
        self, account: str, timestamp: datetime, markers: list[DisplayedMarkerData]
    ) -> None:
        BaseRow.__init__(self, account)
        self.remove_css_class("conversation-row")
        self.set_activatable(False)
        self.type = "displayed_marker"
        self.timestamp = timestamp + timedelta(microseconds=1)

        # Copy markers because we modify the list later
        self._markers = markers.copy()
        self._avatar_stack = AvatarStack(self._account)

        self._timestamp_label = TimeLabel()
        self._timestamp_label.add_css_class("dimmed")

        self.grid.attach(self._timestamp_label, 0, 0, 1, 1)
        self.grid.attach(self._avatar_stack, 1, 0, 1, 1)
        self.grid.set_halign(Gtk.Align.END)
        self.grid.set_column_spacing(6)

        self._update_state()

    def do_unroot(self) -> None:
        BaseRow.do_unroot(self)
        del self._timestamp_label
        del self._avatar_stack

    def has_markers(self) -> bool:
        return bool(self._markers)

    def add_markers(self, markers: list[DisplayedMarkerData]) -> None:
        for marker in markers:
            self._markers.append(marker)
        self._update_state()

    def remove_marker(self, marker: DisplayedMarkerData) -> None:
        for m in self._markers:
            assert m.occupant is not None
            assert marker.occupant is not None
            if m.occupant.id != marker.occupant.id:
                continue

            self._markers.remove(m)
            self._update_state()
            return

    def _update_state(self) -> None:
        self._avatar_stack.set_data(self._markers)
        single_marker = len(self._markers) == 1
        if single_marker:
            marker = self._markers[0]
            self._timestamp_label.set_timestamp(marker.timestamp)

        self._timestamp_label.set_visible(single_marker)
        self.set_visible(bool(self._markers))



'''
This module is in charge of taking care of all the information related to
individual files. Files are identified by the account name and its sid.


>>> print(FilesProp.getFileProp('jabberid', '10'))
None
>>> fp = FilesProp()
Traceback (most recent call last):
    ...
Exception: this class should not be instantiated
>>> print(FilesProp.getAllFileProp())
[]
>>> fp = FilesProp.getNewFileProp('jabberid', '10')
>>> fp2 = FilesProp.getFileProp('jabberid', '10')
>>> fp == fp2
True
'''

from __future__ import annotations

from typing import ClassVar
from typing import Literal

from collections.abc import Callable


class FilesProp:
    _files_props: ClassVar[dict[tuple[str, str], FileProp]] = {}

    def __init__(self):
        raise Exception('this class should not be instantiated')

    @classmethod
    def getNewFileProp(cls, account: str, sid: str) -> FileProp:
        fp = FileProp(account, sid)
        cls.setFileProp(fp, account, sid)
        return fp

    @classmethod
    def getFileProp(cls, account: str, sid: str) -> FileProp | None:
        return cls._files_props.get((account, sid))

    @classmethod
    def getFilePropByAccount(cls, account: str) -> list[FileProp]:
        # Returns a list of file_props in one account
        file_props: list[FileProp] = []
        for account_, sid in cls._files_props:
            if account_ == account:
                file_props.append(cls._files_props[account, sid])
        return file_props

    @classmethod
    def getFilePropByType(cls,
                          type_: Literal['r', 's'],
                          sid: str
                          ) -> FileProp | None:
        # This method should be deleted. Getting fileprop by type and sid is not
        # unique enough. More than one fileprop might have the same type and sid
        files_prop = cls.getAllFileProp()
        for fp in files_prop:
            if fp.type_ == type_ and fp.sid == sid:
                return fp
        return None

    @classmethod
    def getFilePropBySid(cls, sid: str) -> FileProp | None:
        # This method should be deleted. It is kept to make things compatible
        # This method should be replaced and instead get the file_props by
        # account and sid
        files_prop = cls.getAllFileProp()
        for fp in files_prop:
            if fp.sid == sid:
                return fp
        return None

    @classmethod
    def getFilePropByTransportSid(cls,
                                  account: str,
                                  sid: str
                                  ) -> FileProp | None:
        files_prop = cls.getAllFileProp()
        for fp in files_prop:
            if fp.account == account and fp.transport_sid == sid:
                return fp
        return None

    @classmethod
    def getAllFileProp(cls) -> list[FileProp]:
        return list(cls._files_props.values())

    @classmethod
    def setFileProp(cls, fp: FileProp, account: str, sid: str) -> None:
        cls._files_props[account, sid] = fp

    @classmethod
    def deleteFileProp(cls, file_prop: FileProp) -> None:
        files_props = cls._files_props
        acc = sid_ = None
        for key in files_props:
            account, sid = key
            fp = files_props[account, sid]
            if fp is file_prop:
                acc = account
                sid_ = sid
        if acc is not None and sid_ is not None:
            del files_props[acc, sid_]


class FileProp:
    def __init__(self, account: str, sid: str) -> None:
        # Do not instantiate this class directly!
        # Call FilesProp.getNeFileProp instead
        self.streamhosts = []
        self.transferred_size: list[tuple[float, int]] = []
        self.started: bool = False
        self.completed: bool = False
        self.paused: bool = False
        self.stalled: bool = False
        self.connected: bool = False
        self.stopped: bool = False
        self.is_a_proxy: bool = False
        self.proxyhost = None
        self.proxy_sender = None
        self.proxy_receiver = None
        self.streamhost_used = None
        self.failure_cb: Callable[[str], None] | None = None
        self.disconnect_cb = None
        self.continue_cb: Callable[..., None] | None = None
        self.sha_str: str | None = None
        # transfer type: 's' for sending and 'r' for receiving
        self.type_: Literal['r', 's'] | None = None
        self.error: int | None = None
        self.elapsed_time: float = 0  # Elapsed time of the file transfer
        self.last_time: float | None = None
        self.received_len: int | None = None
        self.file_name: str | None = None  # full file path
        self.name: str | None = None
        self.date: str | None = None
        self.desc: str | None = None
        self.offset: int | None = None
        self.sender: str | None = None
        self.receiver: str | None = None
        self.tt_account: str | None = None
        self.size: int | None = None
        self._sid = sid
        self.transport_sid: str | None = None
        self.account = account
        self.mime_type: str | None = None
        self.algo: str | None = None
        self.direction: Literal['<', '>'] | None = None
        self.syn_id: str | None = None
        self.seq: int | None = None
        self.hash_: str | None = None
        self.fd: int | None = None
        # Type of the session, if it is 'jingle' or 'si'
        self.session_type: str | None = None
        self.request_id: str | None = None
        self.proxyhosts = None
        self.dstaddr = None

    def getsid(self) -> str:
        # Getter of the property sid
        return self._sid

    def setsid(self, value: str) -> None:
        # The sid value will change
        # we need to change the in _files_props key as well
        del FilesProp._files_props[self.account, self._sid]
        self._sid = value
        FilesProp._files_props[self.account, self._sid] = self

    sid = property(getsid, setsid)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import dataclasses
import logging
from collections.abc import Sequence

from gi.repository import Gio
from gi.repository import GLib

from gajim.common.dbus.base import DBusService

GetGroupPropertiesT = list[tuple[int, dict[str, GLib.Variant]]]

MENU_NODE_INFO = Gio.DBusNodeInfo.new_for_xml(
    """
<?xml version="1.0" encoding="UTF-8"?>
<node>
    <interface name="com.canonical.dbusmenu">
        <property name="Version" type="u" access="read" />
        <property name="TextDirection" type="s" access="read" />
        <property name="Status" type="s" access="read" />
        <property name="IconThemePath" type="as" access="read" />
        <method name="GetLayout">
            <arg type="i" name="parentId" direction="in" />
            <arg type="i" name="recursionDepth" direction="in" />
            <arg type="as" name="propertyNames" direction="in"  />
            <arg type="u" name="revision" direction="out" />
            <arg type="(ia{sv}av)" name="layout" direction="out" />
        </method>
        <method name="GetGroupProperties">
            <arg type="ai" name="ids" direction="in"/>
            <arg type="as" name="propertyNames" direction="in" />
            <arg type="a(ia{sv})" name="properties" direction="out" />
        </method>
        <method name="GetProperty">
            <arg type="i" name="id" direction="in" />
            <arg type="s" name="name" direction="in" />
            <arg type="v" name="value" direction="out" />
        </method>
        <method name="Event">
            <arg type="i" name="id" direction="in" />
            <arg type="s" name="eventId" direction="in" />
            <arg type="v" name="data" direction="in" />
            <arg type="u" name="timestamp" direction="in" />
        </method>
        <method name="EventGroup">
            <arg type="a(isvu)" name="events" direction="in" />
            <arg type="ai" name="idErrors" direction="out" />
        </method>
        <method name="AboutToShow">
            <arg type="i" name="id" direction="in" />
            <arg type="b" name="needUpdate" direction="out" />
        </method>
        <method name="AboutToShowGroup">
            <arg type="ai" name="ids" direction="in" />
            <arg type="ai" name="updatesNeeded" direction="out" />
            <arg type="ai" name="idErrors" direction="out" />
        </method>
        <signal name="LayoutUpdated">
            <arg type="u" name="revision" direction="out" />
            <arg type="i" name="parent" direction="out" />
        </signal>
        <signal name="ItemsPropertiesUpdated">
            <arg type="a(ia{sv})" name="updatedProps" direction="out" />
            <arg type="a(ias)" name="removedProps" direction="out" />
        </signal>
    </interface>
</node>"""
).interfaces[0]


PROPERTY_NAMES = [
    "children-display",
    "disposition",
    "enabled",
    "icon-name",
    "label",
    "toggle-state",
    "toggle-type",
    "type",
    "visible",
]


log = logging.getLogger("gajim.c.dbus.dbusmenu")


def get_children_from_parent(
    parent_id: int, items: list[DBusMenuItem]
) -> list[DBusMenuItem] | None:
    for item in items:
        if item.children:
            if item.id == parent_id:
                return item.children

            ret = get_children_from_parent(parent_id, item.children)
            if ret is not None:
                return ret

    return None


def remove_invalid_properties(properties: list[str]) -> list[str]:
    return [p for p in properties if p in PROPERTY_NAMES]


@dataclasses.dataclass
class DBusMenu:
    items: list[DBusMenuItem]

    def get_children(self, parent_id: int) -> list[DBusMenuItem]:
        if parent_id == 0:
            return self.items
        children = get_children_from_parent(parent_id, self.items)
        if children is None:
            return []
        return children

    def serialize(
        self, parent_id: int, recursion_depth: int, property_names: list[str]
    ) -> list[GLib.Variant]:
        children = self.get_children(parent_id)
        return [item.serialize(recursion_depth, property_names) for item in children]


@dataclasses.dataclass
class DBusMenuItem:
    id: int
    label: str = ""
    callback: Any = None
    type: str = "standard"
    disposition = "normal"
    enabled: bool = True
    visible: bool = True
    toggle_type: str = ""
    toggle_state: int = -1
    icon_name: str = ""
    children_display: str = ""
    children: list[DBusMenuItem] | None = None

    def flip_toggle_state(self) -> None:
        if self.toggle_state in (0, -1):
            self.toggle_state = 1
        else:
            self.toggle_state = 0

    def serialize_prop(self, name: str) -> GLib.Variant:
        name = name.replace("-", "_")
        value = getattr(self, name)
        if isinstance(value, str):
            return GLib.Variant("s", value)

        if isinstance(value, bool):
            return GLib.Variant("b", value)

        if isinstance(value, int):
            return GLib.Variant("i", value)

        raise ValueError

    def serialize_props(self, property_names: list[str]) -> dict[str, GLib.Variant]:
        result: dict[str, GLib.Variant] = {}

        if not property_names:
            property_names = PROPERTY_NAMES

        for prop in property_names:
            variant = self.serialize_prop(prop)
            result[prop.replace("_", "-")] = variant

        return result

    def serialize(
        self, recursion_depth: int, property_names: list[str]
    ) -> GLib.Variant:
        props = self.serialize_props(property_names)

        children = []
        if recursion_depth > 1 or recursion_depth == -1:
            if self.children:
                children = [
                    item.serialize(recursion_depth - 1, property_names)
                    for item in self.children
                ]

        return GLib.Variant("(ia{sv}av)", (self.id, props, children))


class DBusMenuService(DBusService):
    Status = "normal"
    IconThemePath = ""
    Version = 3
    TextDirection = "ltr"

    def __init__(
        self, object_path: str, session_bus: Gio.DBusConnection, menu: DBusMenu
    ) -> None:
        super().__init__(
            interface_info=MENU_NODE_INFO, object_path=object_path, bus=session_bus
        )

        self._revision = 0
        self._menu = menu

        self.set_menu(menu)

    def set_menu(self, menu: DBusMenu) -> None:
        self._menu = menu
        self._flat_items = self._get_flat_items(self._menu.items)

        self.LayoutUpdated()

    def _get_flat_items(
        self,
        items: list[DBusMenuItem],
        flat_items: dict[int, DBusMenuItem] | None = None,
    ) -> dict[int, DBusMenuItem]:
        if flat_items is None:
            flat_items = {}

        for item in items:
            flat_items[item.id] = item

            if item.children is not None:
                flat_items = self._get_flat_items(item.children, flat_items)

        return flat_items

    def GetLayout(
        self, parent_id: int, recursion_depth: int, property_names: list[str]
    ) -> tuple[int, tuple[int, dict[str, GLib.Variant], list[GLib.Variant]]]:
        property_names = remove_invalid_properties(property_names)

        children = self._menu.serialize(parent_id, recursion_depth, property_names)

        ret = (
            self._revision,
            (0, {"children-display": GLib.Variant("s", "submenu")}, children),
        )

        return ret

    def GetGroupProperties(
        self, item_ids: Sequence[int], property_names: list[str]
    ) -> tuple[GetGroupPropertiesT]:
        ret: GetGroupPropertiesT = []

        property_names = remove_invalid_properties(property_names)

        if not item_ids:
            # Empty means all items
            item_ids = list(self._flat_items.keys())

        for item_id in item_ids:
            if item_id == 0:
                # Sometimes it requests the parent, but the parent has not properties
                continue

            item = self._flat_items.get(item_id)
            if item is None:
                log.warning("Unknown item id: %s", item_id)
                continue

            ret.append((item.id, item.serialize_props(property_names)))

        return (ret,)

    def GetProperty(self, item_id: int, name: str) -> GLib.Variant | None:
        if name not in PROPERTY_NAMES:
            log.warning("Unsupported property: %s", name)
            return None

        item = self._flat_items.get(item_id)
        if item is None:
            log.warning("Unknown item id: %s", item_id)
            return None

        return item.serialize_prop(name)

    def Event(self, item_id: int, event_id: str, _data: Any, _timestamp: Any) -> None:
        if event_id != "clicked":
            return

        item = self._flat_items.get(item_id)
        if item is None:
            log.warning("Unknown item id %s", item_id)
            return

        if item.callback is not None:
            item.callback()

        if item.toggle_type == "checkmark":
            item.flip_toggle_state()
            self.ItemsPropertiesUpdated(item, "toggle-state")

    def EventGroup(
        self, events: list[tuple[int, str, GLib.Variant, int]]
    ) -> tuple[list[int]]:
        not_found: list[int] = []

        for item_id, event_id, _data, _timestamp in events:
            if event_id != "clicked":
                continue

            item = self._flat_items.get(item_id)
            if item is None:
                not_found.append(item_id)
                continue

            if item.callback is not None:
                item.callback()

            if item.toggle_type == "checkmark":
                item.flip_toggle_state()
                self.ItemsPropertiesUpdated(item, "toggle-state")

        return (not_found,)

    def AboutToShow(self, item_id: int) -> tuple[bool]:
        return (False,)

    def AboutToShowGroup(self, item_ids: list[int]) -> tuple[list[Any], list[int]]:
        not_found: list[int] = []

        for item_id in item_ids:
            if item_id not in self._flat_items:
                not_found.append(item_id)
                continue

        return ([], not_found)

    def LayoutUpdated(self, parent: int = 0) -> None:
        self._revision += 1
        self.emit_signal("LayoutUpdated", (self._revision, parent))

    def ItemsPropertiesUpdated(self, item: DBusMenuItem, property_name: str) -> None:
        updated = [(item.id, {property_name: item.serialize_prop(property_name)})]
        self.emit_signal("ItemsPropertiesUpdated", (updated, []))



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Literal
from typing import TypedDict

import logging
from collections.abc import Iterator

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import configpaths
from gajim.common.helpers import dump_json
from gajim.common.helpers import load_json
from gajim.common.i18n import _

log = logging.getLogger("gajim.gtk.shortcut_manager")

GroupsT = Literal["app", "main-win", "input", "input-global"]


class ShortcutManager:
    def __init__(self) -> None:
        self._shortcut_groups = {
            "app": APP_SHORTCUTS,
            "main-win": MAIN_WIN_SHORTCUTS,
            "input": INPUT_SHORTCUTS,
            "input-global": INPUT_GLOBAL_SHORTCUTS,
        }

        self._load_user_shortcuts()

    def store_user_shortcuts(self) -> None:
        user_path = configpaths.get("MY_SHORTCUTS")
        custom_shortcuts = filter(lambda s: s.has_custom_accel(), self.iter_shortcuts())
        user_shortcuts = {s.action_name: s.get_accelerators() for s in custom_shortcuts}
        dump_json(user_path, user_shortcuts)

    def _load_user_shortcuts(self) -> None:
        user_path = configpaths.get("MY_SHORTCUTS")
        user_shortcuts: dict[str, list[str]] = {}
        if user_path.exists():
            log.info("Load user shortcuts")
            user_shortcuts = load_json(user_path, default={})

        for shortcut_group in self._shortcut_groups.values():
            for shortcut in shortcut_group:
                user_accelerators = user_shortcuts.get(shortcut.action_name)
                if user_accelerators is None:
                    continue

                shortcut.set_custom_accelerators(user_accelerators)

    def get_group(self, name: GroupsT) -> GajimShortcutGroup:
        return self._shortcut_groups[name]

    def iter_shortcuts(self) -> Iterator[GajimShortcut]:
        for shortcut_group in self._shortcut_groups.values():
            yield from shortcut_group

    def install_shortcuts(
        self, widget: Gtk.Widget, group_name: list[GroupsT] | GroupsT
    ) -> None:
        if not isinstance(group_name, list):
            group_name = [group_name]

        for name in group_name:
            scope = Gtk.ShortcutScope.LOCAL
            if name.endswith("global"):
                scope = Gtk.ShortcutScope.GLOBAL

            propagation_phase = Gtk.PropagationPhase.BUBBLE
            if name in ("app", "main-win"):
                propagation_phase = Gtk.PropagationPhase.CAPTURE

            controller = Gtk.ShortcutController(
                name=name,
                scope=scope,
                model=self.get_group(name),
                propagation_phase=propagation_phase,
            )
            widget.add_controller(controller)


class AcceleratorDict(TypedDict):
    original: TriggerDict
    custom: TriggerDict


class TriggerDict(TypedDict):
    trigger: Gtk.ShortcutTrigger | None
    accelerators: list[str] | None


class GajimShortcut(Gtk.Shortcut):
    __gtype_name__ = "GajimShortcut"

    def __init__(
        self,
        label: str,
        category: str,
        action_name: str,
        accelerators: list[str] | None = None,
        args: GLib.Variant | None = None,
        allow_rebind: bool = True,
    ):
        Gtk.Shortcut.__init__(self)

        self._label = label
        self._category = category
        self._allow_rebind = allow_rebind
        self._action_name = action_name

        if accelerators is None:
            accelerators = []

        self._accelerators: AcceleratorDict = {
            "original": {"trigger": None, "accelerators": None},
            "custom": {"trigger": None, "accelerators": None},
        }

        self._set_accelerators("original", accelerators)

        shortcut_action = Gtk.ShortcutAction.parse_string(f"action({action_name})")
        self.set_action(shortcut_action)

        if args is not None:
            self.set_arguments(args)

    @GObject.Property(type=str, flags=GObject.ParamFlags.READABLE)
    def label(self) -> str:
        return self._label

    @GObject.Property(type=str, flags=GObject.ParamFlags.READABLE)
    def category(self) -> str:
        return self._category

    @GObject.Property(type=bool, default=True, flags=GObject.ParamFlags.READABLE)
    def allow_rebind(self) -> bool:
        return self._allow_rebind

    @GObject.Property(type=str, flags=GObject.ParamFlags.READABLE)
    def action_name(self) -> str:
        return self._action_name

    def reset(self) -> None:
        self._accelerators["custom"]["trigger"] = None
        self._accelerators["custom"]["accelerators"] = None
        self.set_trigger(self._accelerators["original"]["trigger"])

    def get_accel_label(self) -> str:
        display = Gdk.Display.get_default()
        assert display is not None

        for name in ("custom", "original"):
            trigger = self._accelerators[name]["trigger"]
            if trigger is not None:
                if isinstance(trigger, Gtk.NeverTrigger):
                    return _("Disabled")
                return trigger.to_label(display)

        return _("Disabled")

    def set_custom_accelerators(self, accelerators: list[str] | None) -> None:
        self._set_accelerators("custom", accelerators)

    def has_custom_accel(self) -> bool:
        custom_trigger = self._accelerators["custom"]["trigger"]
        if custom_trigger is None:
            return False

        orig_trigger = self._accelerators["original"]["trigger"]
        if orig_trigger is None:
            return True

        return not orig_trigger.equal(custom_trigger)

    def get_accelerators(self) -> list[str]:
        name = "custom" if self.has_custom_accel() else "original"
        return self._accelerators[name]["accelerators"] or []

    def _set_accelerators(self, name: str, accelerators: list[str] | None) -> None:
        if not accelerators:
            trigger = Gtk.NeverTrigger.get()
            self._accelerators[name]["accelerators"] = None
            self._accelerators[name]["trigger"] = trigger
            self.set_trigger(trigger)
            return

        # Support max two accelerators
        accelerators = accelerators[:2]

        try:
            trigger = Gtk.ShortcutTrigger.parse_string("|".join(accelerators))
        except Exception:
            trigger = None

        if trigger is None:
            log.warning(
                "Unable to create shortcut trigger for: %s %s", name, accelerators
            )
            return

        self._accelerators[name]["accelerators"] = accelerators
        self._accelerators[name]["trigger"] = trigger
        self.set_trigger(trigger)


class GajimShortcutGroup(Gio.ListStore):
    def __init__(self, name: str, shortcuts: list[GajimShortcut]) -> None:
        super().__init__(item_type=GajimShortcut)

        self._name = name
        self._shortcuts: dict[str, GajimShortcut] = {}

        for shortcut in shortcuts:
            self._shortcuts[shortcut.action_name] = shortcut
            self.append(shortcut)

    @GObject.Property(type=str, flags=GObject.ParamFlags.READABLE)
    def name(self) -> str:
        return self._name

    def get_for_action(self, action_name: str) -> GajimShortcut:
        return self._shortcuts[action_name]


APP_SHORTCUTS = GajimShortcutGroup(
    "app",
    [
        GajimShortcut(
            label=_("Start / Join Chat"),
            category="general",
            accelerators=["<Primary>N"],
            action_name="app.start-chat",
            args=GLib.Variant("as", ["", ""]),
        ),
        GajimShortcut(
            label=_("Create New Group Chat"),
            category="general",
            accelerators=["<Primary>G"],
            action_name="app.create-groupchat",
            args=GLib.Variant("s", ""),
        ),
        GajimShortcut(
            label=_("Preferences"),
            category="general",
            accelerators=["<Primary>P"],
            action_name="app.preferences",
        ),
        GajimShortcut(
            label=_("Plugins"),
            category="general",
            accelerators=["<Primary>E"],
            action_name="app.plugins",
        ),
        GajimShortcut(
            label=_("Manage Shortcuts"),
            category="general",
            accelerators=["<Primary>question"],
            action_name="app.shortcuts",
        ),
        GajimShortcut(
            label=_("Debug Console"),
            category="general",
            accelerators=["<Primary><Shift>X"],
            action_name="app.xml-console",
        ),
        GajimShortcut(
            label=_("Quit Gajim"),
            category="general",
            accelerators=["<Primary>Q"],
            action_name="app.quit",
        ),
    ],
)

MAIN_WIN_SHORTCUTS = GajimShortcutGroup(
    "main-win",
    [
        GajimShortcut(
            label=_("Increase Font Size"),
            category="general",
            accelerators=["<Primary>plus", "<Primary>KP_Add"],
            action_name="win.increase-app-font-size",
        ),
        GajimShortcut(
            label=_("Decrease Font Size"),
            category="general",
            accelerators=["<Primary>minus", "<Primary>KP_Subtract"],
            action_name="win.decrease-app-font-size",
        ),
        GajimShortcut(
            label=_("Reset Font Size"),
            category="general",
            accelerators=["<Primary>0", "<Primary>KP_0"],
            action_name="win.reset-app-font-size",
        ),
        GajimShortcut(
            label=_("Focus Search"),
            category="chats",
            accelerators=["<Primary>K"],
            action_name="win.focus-search",
        ),
        GajimShortcut(
            label=_("Search Chat History"),
            category="chats",
            accelerators=["<Primary>F"],
            action_name="win.search-history",
        ),
        GajimShortcut(
            label=_("Contact Details"),
            category="chats",
            accelerators=["<Primary>I"],
            action_name="win.show-contact-info",
        ),
        GajimShortcut(
            label=_("Change Nickname"),
            category="chats",
            accelerators=["<Primary><Shift>N"],
            action_name="win.change-nickname",
        ),
        GajimShortcut(
            label=_("Change Subject"),
            category="chats",
            accelerators=["<Primary><Shift>S"],
            action_name="win.change-subject",
        ),
        GajimShortcut(
            label="Escape",
            category="chats",
            accelerators=["Escape"],
            action_name="win.escape",
            allow_rebind=False,
        ),
        GajimShortcut(
            label=_("Close Chat"),
            category="chats",
            accelerators=["<Primary>W"],
            action_name="win.close-chat",
        ),
        GajimShortcut(
            label=_("Restore Closed Chat"),
            category="chats",
            accelerators=["<Primary><Shift>W"],
            action_name="win.restore-chat",
        ),
        GajimShortcut(
            label=_("Toggle Chat List"),
            category="chats",
            accelerators=["<Primary>R"],
            action_name="win.chat-list-visible",
        ),
        GajimShortcut(
            label=_("Switch to Next Chat"),
            category="chats",
            accelerators=["<Primary>Page_Down"],
            action_name="win.switch-next-chat",
        ),
        GajimShortcut(
            label=_("Switch to Previous Chat"),
            category="chats",
            accelerators=["<Primary>Page_Up"],
            action_name="win.switch-prev-chat",
        ),
        GajimShortcut(
            label=_("Switch to Next Unread Chat"),
            category="chats",
            accelerators=["<Primary>Tab"],
            action_name="win.switch-next-unread-chat",
        ),
        GajimShortcut(
            label=_("Switch to Previous Unread Chat"),
            category="chats",
            accelerators=["<Primary>ISO_Left_Tab"],
            action_name="win.switch-prev-unread-chat",
        ),
        GajimShortcut(
            label=_("Switch to Chat 1"),
            category="chats",
            accelerators=["<Alt>1", "<Alt>KP_1"],
            action_name="win.switch-chat-1",
        ),
        GajimShortcut(
            label=_("Switch to Chat 2"),
            category="chats",
            accelerators=["<Alt>2", "<Alt>KP_2"],
            action_name="win.switch-chat-2",
        ),
        GajimShortcut(
            label=_("Switch to Chat 3"),
            category="chats",
            accelerators=["<Alt>3", "<Alt>KP_3"],
            action_name="win.switch-chat-3",
        ),
        GajimShortcut(
            label=_("Switch to Chat 4"),
            category="chats",
            accelerators=["<Alt>4", "<Alt>KP_4"],
            action_name="win.switch-chat-4",
        ),
        GajimShortcut(
            label=_("Switch to Chat 5"),
            category="chats",
            accelerators=["<Alt>5", "<Alt>KP_5"],
            action_name="win.switch-chat-5",
        ),
        GajimShortcut(
            label=_("Switch to Chat 6"),
            category="chats",
            accelerators=["<Alt>6", "<Alt>KP_6"],
            action_name="win.switch-chat-6",
        ),
        GajimShortcut(
            label=_("Switch to Chat 7"),
            category="chats",
            accelerators=["<Alt>7", "<Alt>KP_7"],
            action_name="win.switch-chat-7",
        ),
        GajimShortcut(
            label=_("Switch to Chat 8"),
            category="chats",
            accelerators=["<Alt>8", "<Alt>KP_8"],
            action_name="win.switch-chat-8",
        ),
        GajimShortcut(
            label=_("Switch to Chat 9"),
            category="chats",
            accelerators=["<Alt>9", "<Alt>KP_9"],
            action_name="win.switch-chat-9",
        ),
        GajimShortcut(
            label=_("Switch to Workspace 1"),
            category="chats",
            accelerators=["<Primary>1", "<Primary>KP_1"],
            action_name="win.switch-workspace-1",
        ),
        GajimShortcut(
            label=_("Switch to Workspace 2"),
            category="chats",
            accelerators=["<Primary>2", "<Primary>KP_2"],
            action_name="win.switch-workspace-2",
        ),
        GajimShortcut(
            label=_("Switch to Workspace 3"),
            category="chats",
            accelerators=["<Primary>3", "<Primary>KP_3"],
            action_name="win.switch-workspace-3",
        ),
        GajimShortcut(
            label=_("Switch to Workspace 4"),
            category="chats",
            accelerators=["<Primary>4", "<Primary>KP_4"],
            action_name="win.switch-workspace-4",
        ),
        GajimShortcut(
            label=_("Switch to Workspace 5"),
            category="chats",
            accelerators=["<Primary>5", "<Primary>KP_5"],
            action_name="win.switch-workspace-5",
        ),
        GajimShortcut(
            label=_("Switch to Workspace 6"),
            category="chats",
            accelerators=["<Primary>6", "<Primary>KP_6"],
            action_name="win.switch-workspace-6",
        ),
        GajimShortcut(
            label=_("Switch to Workspace 7"),
            category="chats",
            accelerators=["<Primary>7", "<Primary>KP_7"],
            action_name="win.switch-workspace-7",
        ),
        GajimShortcut(
            label=_("Switch to Workspace 8"),
            category="chats",
            accelerators=["<Primary>8", "<Primary>KP_8"],
            action_name="win.switch-workspace-8",
        ),
        GajimShortcut(
            label=_("Switch to Workspace 9"),
            category="chats",
            accelerators=["<Primary>9", "<Primary>KP_9"],
            action_name="win.switch-workspace-9",
        ),
        GajimShortcut(
            label=_("Scroll Up"),
            category="messages",
            accelerators=["<Shift>Page_Up"],
            action_name="win.scroll-view-up",
        ),
        GajimShortcut(
            label=_("Scroll Down"),
            category="messages",
            accelerators=["<Shift>Page_Down"],
            action_name="win.scroll-view-down",
        ),
        GajimShortcut(
            label=_("Quote Previous Message"),
            category="messages",
            accelerators=["<Primary><Shift>Up"],
            action_name="win.quote-prev",
        ),
        GajimShortcut(
            label=_("Quote Next Message"),
            category="messages",
            accelerators=["<Primary><Shift>Down"],
            action_name="win.quote-next",
        ),
    ],
)


INPUT_SHORTCUTS = GajimShortcutGroup(
    "input",
    [
        GajimShortcut(
            label=_("Clear Input"),
            category="messages",
            accelerators=["<Primary>U"],
            action_name="win.input-clear",
        ),
        GajimShortcut(
            label=_("Correct Message"),
            category="messages",
            accelerators=["<Primary>Up"],
            action_name="win.correct-message",
        ),
        GajimShortcut(
            label=_("Format Text Bold"),
            category="messages",
            accelerators=["<Primary>B"],
            action_name="win.input-bold",
        ),
        GajimShortcut(
            label=_("Format Text Italic"),
            category="messages",
            accelerators=None,
            action_name="win.input-italic",
        ),
        GajimShortcut(
            label=_("Format Text Strikethrough"),
            category="messages",
            accelerators=None,
            action_name="win.input-strike",
        ),
        GajimShortcut(
            label=_("Open Emoji Chooser"),
            category="messages",
            accelerators=None,
            action_name="win.input-emoji",
        ),
        GajimShortcut(
            label=_("Paste as Quote"),
            category="messages",
            accelerators=None,
            action_name="win.input-paste-as-quote",
        ),
        GajimShortcut(
            label=_("Paste as Code Block"),
            category="messages",
            accelerators=None,
            action_name="win.input-paste-as-code-block",
        ),
    ],
)


INPUT_GLOBAL_SHORTCUTS = GajimShortcutGroup(
    "input-global",
    [
        GajimShortcut(
            label=_("Focus Input"),
            category="messages",
            accelerators=None,
            action_name="win.input-focus",
        ),
    ],
)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import datetime as dt

from gi.repository import GLib
from nbxmpp.modules.vcard4 import TzProperty
from nbxmpp.modules.vcard4 import VCard

from gajim.common.iana import get_zone_data

FIRST_UTC_DATETIME = dt.datetime(1970, 1, 1, tzinfo=dt.UTC)
FIRST_LOCAL_DATETIME = FIRST_UTC_DATETIME.astimezone()


def convert_epoch_to_local_datetime(utc_timestamp: float) -> dt.datetime:
    utc = dt.datetime.fromtimestamp(utc_timestamp, tz=dt.UTC)
    return utc.astimezone()


def utc_now() -> dt.datetime:
    return dt.datetime.now(dt.UTC)


def get_start_of_day(datetime: dt.datetime) -> dt.datetime:
    return datetime.replace(hour=0, minute=0, second=0, microsecond=0)


def get_local_timezone() -> str | None:
    timezone = GLib.TimeZone.new_local().get_identifier() or None
    if timezone is None:
        return None

    zone_data = get_zone_data(timezone)
    if zone_data is None:
        return None

    return zone_data.key


def get_timezone_from_vcard(vcard: VCard) -> str | None:
    for prop in vcard.get_properties():
        match prop:
            case TzProperty():
                return prop.value or None

            case _:
                pass



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Final

import logging

from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.completion.base import BaseCompletionListItem
from gajim.gtk.completion.base import BaseCompletionProvider
from gajim.gtk.completion.base import BaseCompletionViewItem

log = logging.getLogger("gajim.gtk.completion.commands")


MAX_COMPLETION_ENTRIES = 10


class CommandsCompletionListItem(BaseCompletionListItem, GObject.Object):
    __gtype_name__ = "CommandsCompletionListItem"

    command = GObject.Property(type=str)
    usage = GObject.Property(type=str)

    def get_text(self) -> str:
        return f"/{self.command} "


class CommandsCompletionViewItem(
    BaseCompletionViewItem[CommandsCompletionListItem], Gtk.Box
):
    __gtype_name__ = "CommandsCompletionViewItem"
    css_class = "command-completion"

    def __init__(self) -> None:
        super().__init__()
        Gtk.Box.__init__(self)

        self._label = Gtk.Label()

        self.append(self._label)

    def bind(self, obj: CommandsCompletionListItem) -> None:
        bind_spec = [
            ("usage", self._label, "label"),
        ]

        for source_prop, widget, target_prop in bind_spec:
            bind = obj.bind_property(
                source_prop, widget, target_prop, GObject.BindingFlags.SYNC_CREATE
            )
            self._bindings.append(bind)

    def unbind(self) -> None:
        for bind in self._bindings:
            bind.unbind()
        self._bindings.clear()

    def do_unroot(self) -> None:
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)


class CommandsCompletionProvider(BaseCompletionProvider):
    trigger_char: Final = "/"
    name = _("Commands")

    def __init__(self) -> None:
        self._list_store = Gio.ListStore(item_type=CommandsCompletionListItem)

        for command, usage in app.commands.get_commands("chat"):  # TODO
            self._list_store.append(
                CommandsCompletionListItem(command=command, usage=f"/{command} {usage}")
            )

        expression = Gtk.PropertyExpression.new(
            CommandsCompletionListItem, None, "command"
        )

        self._string_filter = Gtk.StringFilter(expression=expression)

        filter_model = Gtk.FilterListModel(
            model=self._list_store, filter=self._string_filter
        )
        self._model = Gtk.SliceListModel(
            model=filter_model, size=MAX_COMPLETION_ENTRIES
        )

    def get_model(
        self,
    ) -> tuple[
        Gio.ListModel[CommandsCompletionListItem], type[CommandsCompletionViewItem]
    ]:
        return self._model, CommandsCompletionViewItem

    def check(self, candidate: str, start_iter: Gtk.TextIter) -> bool:
        if not candidate.startswith(self.trigger_char) or candidate.startswith(
            self.trigger_char, 1, 2
        ):
            # Check for '/' at the beginning, but ignore '//'
            return False

        return start_iter.get_offset() == 0

    def populate(self, candidate: str, contact: Any) -> bool:
        candidate = candidate.lstrip(self.trigger_char)
        self._string_filter.set_search(candidate)
        return self._model.get_n_items() > 0



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0292: vCard4 Over XMPP

from __future__ import annotations

from typing import Any
from typing import cast

import datetime as dt
import inspect
import weakref

from nbxmpp.errors import StanzaError
from nbxmpp.errors import TimeoutStanzaError
from nbxmpp.modules.vcard4 import TzProperty
from nbxmpp.modules.vcard4 import VCard
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.structs import MessageProperties
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import types
from gajim.common.events import TimezoneChanged
from gajim.common.events import VCard4Received
from gajim.common.modules.base import BaseModule
from gajim.common.modules.util import event_node
from gajim.common.util.datetime import get_local_timezone
from gajim.common.util.datetime import get_timezone_from_vcard
from gajim.common.util.datetime import utc_now


class VCard4(BaseModule):

    _nbxmpp_extends = 'VCard4'
    _nbxmpp_methods = [
        'set_vcard',
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)
        self._vcard_cache: dict[JID, tuple[dt.datetime, VCard]] = {}
        self._own_vcard: VCard = VCard()
        self._register_pubsub_handler(self._vcard_event_received)

    @event_node(Namespace.VCARD4_PUBSUB)
    def _vcard_event_received(
        self,
        _client: types.NBXMPPClient,
        _stanza: Message,
        properties: MessageProperties
        ) -> None:

        assert properties.jid is not None
        if not properties.jid.bare_match(self._get_own_bare_jid()):
            self._log.info(
                "Ignore VCard4 event not from our account: %s", properties.jid)
            return

        assert properties.pubsub_event is not None
        if (properties.pubsub_event.deleted
                or properties.pubsub_event.purged
                or properties.pubsub_event.retracted):
            self._log.info('VCard4 node deleted/purged/retracted')
            self._own_vcard = VCard()

        else:
            self._log.info("Received VCard4 event from %s", properties.jid)
            self._own_vcard = cast(VCard, properties.pubsub_event.data)

        app.ged.raise_event(VCard4Received(self._account, self._own_vcard))

        self._check_for_timezone_change()

    def _check_for_timezone_change(self) -> None:
        if not app.settings.get_account_setting(self._account, "update_timezone"):
            return

        vcard_timezone = get_timezone_from_vcard(self._own_vcard)
        local_timezone = get_local_timezone()
        if local_timezone is None:
            self._log.warning("Unable to determine timezone")
            return

        if vcard_timezone == local_timezone:
            return

        self._log.info("Timezone change detected, vcard: %s, local: %s",
                       vcard_timezone, local_timezone)

        app.ged.raise_event(
            TimezoneChanged(
                self._account,
                vcard=vcard_timezone,
                local=local_timezone,
            )
        )

    def update_timezone(self) -> None:
        local_timezone = get_local_timezone()
        if local_timezone is None:
            self._log.warning("Unable to determine timezone for update")
            return

        for prop in self._own_vcard.get_properties():
            if isinstance(prop, TzProperty):
                self._own_vcard.remove_property(prop)

        self._own_vcard.add_property("tz", value_type="text", value=local_timezone)
        self._log.info("Update timezone to: %s", local_timezone)
        self.set_vcard(self._own_vcard)

    def is_timezone_published(self) -> bool:
        return get_timezone_from_vcard(self._own_vcard) is not None

    def subscribe_to_node(self) -> None:
        self._log.info("Subscribe to node")

        jid = self._get_own_bare_jid()
        # JID is necessary because ejabberd servers react weird to subscribe
        # to own nodes if 'to' attribute is not set
        self._client.get_module('PubSub').subscribe(Namespace.VCARD4_PUBSUB, jid)

        self._nbxmpp('VCard4').request_vcard(
            jid,
            timeout=10,
            callback=self._on_vcard_received,
            user_data=(jid, None),
        )

    def get_own_vcard(self) -> VCard:
        return self._own_vcard

    def request_vcard(
        self, jid: JID, callback: Any, max_cache_seconds: int = 0
    ) -> None | VCard:
        if max_cache_seconds > 0:
            cached_result = self._vcard_cache.get(jid)
            if cached_result is not None:
                request_dt, vcard = cached_result
                if dt.timedelta(seconds=max_cache_seconds) > utc_now() - request_dt:
                    return vcard

        if inspect.ismethod(callback):
            weak_callable = weakref.WeakMethod(callback)
        elif inspect.isfunction(callback):
            weak_callable = weakref.ref(callback)
        else:
            raise TypeError('Unknown callback type: %s' % callback)

        self._nbxmpp('VCard4').request_vcard(
            jid,
            timeout=10,
            callback=self._on_vcard_received,
            user_data=(jid, weak_callable),
        )

    def _on_vcard_received(self, task: Task) -> None:
        try:
            vcard = cast(VCard | None, task.finish())
        except (StanzaError, TimeoutStanzaError) as err:
            self._log.info('Error loading VCard: %s', err)
            vcard = None

        jid, weak_callable = task.get_user_data()

        if vcard is None:
            vcard = VCard()
        else:
            self._log.info('Received VCard for %s', jid)
            self._vcard_cache[jid] = (utc_now(), vcard)

        if jid.bare_match(self._get_own_bare_jid()):
            self._own_vcard = vcard
            app.ged.raise_event(VCard4Received(self._account, self._own_vcard))
            self._check_for_timezone_change()

        if weak_callable is None:
            return

        callback = weak_callable()
        if callback is None:
            return

        callback(jid, vcard)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import logging

import nbxmpp
from nbxmpp.namespaces import Namespace

from gajim.common import app
from gajim.common import types
from gajim.common.jingle_transport import TransportType

# from gajim.common.socks5 import Socks5ReceiverClient
# from gajim.common.socks5 import Socks5SenderClient

if TYPE_CHECKING:
    from gajim.common.jingle_ft import JingleFileTransfer

log = logging.getLogger('gajim.c.jingle_ftstates')


class JingleFileTransferStates:
    '''
    This class implements the state machine design pattern
    '''

    def __init__(self, jingleft: JingleFileTransfer) -> None:
        self.jft = jingleft

    def action(self, args: dict[str, Any] | None = None) -> None:
        '''
        This method MUST be overridden by a subclass
        '''
        raise NotImplementedError('This is an abstract method!')


class StateInitialized(JingleFileTransferStates):
    '''
    This state initializes the file transfer
    '''

    def action(self, args: dict[str, Any] | None = None) -> None:
        if self.jft.weinitiate:
            # update connection's fileprops
            self.jft._listen_host()
            # Listen on configured port for file transfer
        else:
            # Connect to the candidate host, on success call on_connect method
            app.socks5queue.connect_to_hosts(
                self.jft.session.connection.name,
                self.jft.file_props.transport_sid, self.jft.on_connect,
                self.jft._on_connect_error)


class StateCandSent(JingleFileTransferStates):
    '''
    This state sends our nominated candidate
    '''

    def _send_candidate(self, args: dict[str, Any]) -> None:
        if 'candError' in args:
            self.jft.nominated_cand['our-cand'] = False
            self.jft.send_error_candidate()
            return
        # Send candidate used
        streamhost = args['streamhost']
        self.jft.nominated_cand['our-cand'] = streamhost
        content = nbxmpp.Node('content')
        content.setAttr('creator', 'initiator')
        content.setAttr('name', self.jft.name)
        transport = nbxmpp.Node('transport')
        transport.setNamespace(Namespace.JINGLE_BYTESTREAM)
        transport.setAttr('sid', self.jft.transport.sid)
        candidateused = nbxmpp.Node('candidate-used')
        candidateused.setAttr('cid', streamhost['candidate_id'])
        transport.addChild(node=candidateused)
        content.addChild(node=transport)
        self.jft.session.send_transport_info(content)

    def action(self, args: dict[str, Any] | None = None) -> None:
        self._send_candidate(args)


class StateCandReceived(JingleFileTransferStates):
    '''
    This state happens when we receive a candidate.
    It takes the arguments: canError if we receive a candidate-error
    '''

    def _recv_candidate(self, args: dict[str, Any]) -> None:
        if 'candError' in args:
            return
        content = args['content']
        streamhost_cid = content.getTag('transport').getTag('candidate-used').\
            getAttr('cid')
        streamhost_used = None
        for cand in self.jft.transport.candidates:
            if cand['candidate_id'] == streamhost_cid:
                streamhost_used = cand
                break
        if streamhost_used is None:
            log.info('unknown streamhost')
            return
        # We save the candidate nominated by peer
        self.jft.nominated_cand['peer-cand'] = streamhost_used

    def action(self, args: dict[str, Any] | None = None) -> None:
        self._recv_candidate(args)


class StateCandSentAndRecv(StateCandSent, StateCandReceived):
    '''
    This state happens when we have received and sent the candidates.
    It takes the boolean argument: sendCand in order to decide whether
    we should execute the action of when we receive or send a candidate.
    '''

    def action(self, args: dict[str, Any] | None = None) -> None:
        if args['sendCand']:
            self._send_candidate(args)
        else:
            self._recv_candidate(args)


class StateTransportReplace(JingleFileTransferStates):
    '''
    This state initiates transport replace
    '''

    def action(self, args: dict[str, Any] | None = None) -> None:
        self.jft.session.transport_replace()


class StateTransfering(JingleFileTransferStates):
    '''
    This state will start the transfer depending on the type of transport
    we have.
    '''

    def _start_ibb_transfer(self, con: types.Client) -> None:
        self.jft.file_props.transport_sid = self.jft.transport.sid
        fp = open(self.jft.file_props.file_name, 'rb')
        con.get_module('IBB').send_open(self.jft.session.peerjid,
                                        self.jft.file_props.sid,
                                        fp)

    def _start_sock5_transfer(self) -> None:
        # It tells whether we start the transfer as client or server
        mode = None
        if self.jft.is_our_candidate_used():
            mode = 'client'
            streamhost_used = self.jft.nominated_cand['our-cand']
            app.socks5queue.remove_server(self.jft.file_props.transport_sid)
        else:
            mode = 'server'
            streamhost_used = self.jft.nominated_cand['peer-cand']
            app.socks5queue.remove_client(self.jft.file_props.transport_sid)
            app.socks5queue.remove_other_servers(streamhost_used['host'])
        if streamhost_used['type'] == 'proxy':
            self.jft.file_props.is_a_proxy = True
            if self.jft.file_props.type_ == 's' and self.jft.weinitiate:
                self.jft.file_props.proxy_sender = streamhost_used['initiator']
                self.jft.file_props.proxy_receiver = streamhost_used['target']
            else:
                self.jft.file_props.proxy_sender = streamhost_used['target']
                self.jft.file_props.proxy_receiver = streamhost_used[
                    'initiator']
            if self.jft.file_props.type_ == 's':
                s = app.socks5queue.senders
                for v in s.values():
                    if (v.host == streamhost_used['host'] and
                            v.connected):
                        return
            elif self.jft.file_props.type_ == 'r':
                r = app.socks5queue.readers
                for v in r.values():
                    if (v.host == streamhost_used['host'] and
                            v.connected):
                        return
            else:
                raise TypeError
            self.jft.file_props.streamhost_used = True
            streamhost_used['sid'] = self.jft.file_props.transport_sid
            self.jft.file_props.streamhosts = []
            self.jft.file_props.streamhosts.append(streamhost_used)
            self.jft.file_props.proxyhosts = []
            self.jft.file_props.proxyhosts.append(streamhost_used)
            if self.jft.file_props.type_ == 's':
                app.socks5queue.idx += 1
                idx = app.socks5queue.idx
                sockobj = Socks5SenderClient(app.idlequeue, idx,
                                             app.socks5queue, _sock=None,
                                             host=str(streamhost_used['host']),
                                             port=int(streamhost_used['port']),
                                             connected=False,
                                             file_props=self.jft.file_props)
            else:
                sockobj = Socks5ReceiverClient(
                    app.idlequeue, streamhost_used,
                    transport_sid=self.jft.file_props.transport_sid,
                    file_props=self.jft.file_props)
            sockobj.proxy = True
            sockobj.streamhost = streamhost_used
            app.socks5queue.add_sockobj(self.jft.session.connection.name,
                                        sockobj)
            streamhost_used['idx'] = sockobj.queue_idx
            # If we offered the nominated candidate used, we activate
            # the proxy
            if not self.jft.is_our_candidate_used():
                app.socks5queue.on_success[self.jft.file_props.transport_sid]\
                    = self.jft.transport._on_proxy_auth_ok
            # TODO: add on failure
        else:
            app.socks5queue.send_file(self.jft.file_props,
                                      self.jft.session.connection.name, mode)

    def action(self, args: dict[str, Any] | None = None) -> None:
        if self.jft.transport.type_ == TransportType.IBB:
            self._start_ibb_transfer(self.jft.session.connection)
        elif self.jft.transport.type_ == TransportType.SOCKS5:
            self._start_sock5_transfer()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Final

import logging

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common import ged
from gajim.common.const import AvatarSize
from gajim.common.events import MessageReceived
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant

from gajim.gtk.completion.base import BaseCompletionListItem
from gajim.gtk.completion.base import BaseCompletionProvider
from gajim.gtk.completion.base import BaseCompletionViewItem

log = logging.getLogger("gajim.gtk.completion.nickname")


MAX_COMPLETION_ENTRIES = 10


class NicknameCompletionListItem(BaseCompletionListItem, GObject.Object):
    __gtype_name__ = "NicknameCompletionListItem"

    nickname = GObject.Property(type=str)
    avatar = GObject.Property(type=Gdk.Texture)

    def get_text(self) -> str:
        return f"{self.nickname}{app.settings.get('gc_refer_to_nick_char')} "


class NicknameCompletionViewItem(
    BaseCompletionViewItem[NicknameCompletionListItem], Gtk.Box
):
    __gtype_name__ = "NicknameCompletionViewItem"
    css_class = "nickname-completion"

    def __init__(self) -> None:
        super().__init__()
        Gtk.Box.__init__(self, spacing=6)
        self.set_size_request(200, -1)

        self._label = Gtk.Label()
        self._image = Gtk.Image(pixel_size=AvatarSize.SMALL)
        self.append(self._image)
        self.append(self._label)

    def bind(self, obj: NicknameCompletionListItem) -> None:
        bind_spec = [
            ("nickname", self._label, "label"),
            ("avatar", self._image, "paintable"),
        ]

        for source_prop, widget, target_prop in bind_spec:
            bind = obj.bind_property(
                source_prop, widget, target_prop, GObject.BindingFlags.SYNC_CREATE
            )
            self._bindings.append(bind)

    def unbind(self) -> None:
        for bind in self._bindings:
            bind.unbind()
        self._bindings.clear()

    def do_unroot(self) -> None:
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)


class NicknameCompletionProvider(BaseCompletionProvider):
    trigger_char: Final = "@"
    name = _("Participants")

    def __init__(self) -> None:
        self._list_store = Gio.ListStore(item_type=NicknameCompletionListItem)
        self._contact: GroupchatContact | None = None

        app.ged.register_event_handler(
            "message-received", ged.GUI2, self._on_message_received
        )

        expression = Gtk.PropertyExpression.new(
            NicknameCompletionListItem, None, "nickname"
        )

        self._string_filter = Gtk.StringFilter(expression=expression)

        filter_model = Gtk.FilterListModel(
            model=self._list_store, filter=self._string_filter
        )
        self._model = Gtk.SliceListModel(
            model=filter_model, size=MAX_COMPLETION_ENTRIES
        )

    def get_model(
        self,
    ) -> tuple[
        Gio.ListModel[NicknameCompletionListItem], type[NicknameCompletionViewItem]
    ]:
        return self._model, NicknameCompletionViewItem

    def check(self, candidate: str, start_iter: Gtk.TextIter) -> bool:
        return candidate.startswith(self.trigger_char)

    def populate(self, candidate: str, contact: Any) -> bool:
        if not isinstance(contact, GroupchatContact):
            return False

        scale = app.window.get_scale_factor()
        if self._contact is not contact:
            # New contact, regenerate suggestions
            self._list_store.remove_all()
            for suggested_contact in self._generate_suggestions(contact):
                self._list_store.append(
                    NicknameCompletionListItem(
                        nickname=suggested_contact.name,
                        avatar=suggested_contact.get_avatar(
                            AvatarSize.SMALL, scale, add_show=False
                        ),
                    )
                )
            self._contact = contact

        candidate = candidate.lstrip(self.trigger_char)
        self._string_filter.set_search(candidate)
        return self._model.get_n_items() > 0

    def _generate_suggestions(
        self, contact: GroupchatContact
    ) -> list[GroupchatParticipant]:
        # Get recent nicknames from DB. This enables us to suggest
        # nicknames even if no message arrived since Gajim was started.
        recent_participants = list(
            map(
                contact.get_resource,
                app.storage.archive.get_recent_muc_nicks(contact.account, contact.jid),
            )
        )

        current_participants = list(contact.get_participants())
        current_participants.sort(key=lambda c: c.name.lower())

        suggestions = recent_participants + current_participants
        # deduplicate
        return list(dict.fromkeys(suggestions))

    def _on_message_received(self, event: MessageReceived) -> None:
        if self._contact is None:
            return

        if event.jid != self._contact.jid:
            return

        # This will trigger generating new suggestions
        self._contact = None



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from pathlib import Path

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk

from gajim.common import app

from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string


@Gtk.Template.from_string(string=get_ui_string("preview/file_control_buttons.ui"))
class FileControlButtons(Gtk.Box, SignalManager):
    __gtype_name__ = "FileControlButtons"

    _file_name_label: Gtk.Label = Gtk.Template.Child()
    _file_size_label: Gtk.Label = Gtk.Template.Child()
    _open_folder_button: Gtk.Button = Gtk.Template.Child()
    _save_as_button: Gtk.Button = Gtk.Template.Child()

    def __init__(
        self, path: Path | None = None, file_name: str = "", file_size: int = 0
    ) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        pointer_cursor = Gdk.Cursor.new_from_name("pointer")
        self._save_as_button.set_cursor(pointer_cursor)
        self._open_folder_button.set_cursor(pointer_cursor)

        if path is not None:
            self.set_path(path)

        self.set_file_name(file_name)
        self.set_file_size(file_size)

    def set_path(self, path: Path) -> None:
        self._open_folder_button.set_action_target_value(GLib.Variant("s", str(path)))
        self._save_as_button.set_action_target_value(GLib.Variant("s", str(path)))
        self.set_buttons_visible(True)

    def set_file_name(self, file_name: str):
        self._file_name_label.set_text(file_name)

    def set_file_size(self, file_size: int):
        unit = GLib.FormatSizeFlags.DEFAULT
        if app.settings.get("use_kib_mib"):
            unit = GLib.FormatSizeFlags.IEC_UNITS

        file_size_string = ""
        if file_size > 0:
            file_size_string = GLib.format_size_full(file_size, unit)

        self._file_size_label.set_text(file_size_string)
        self._file_size_label.set_visible(bool(file_size_string))

    def set_buttons_visible(self, visible: bool) -> None:
        self._open_folder_button.set_visible(visible)
        self._save_as_button.set_visible(visible)

    def do_unroot(self) -> None:
        Gtk.Box.do_unroot(self)
        self._disconnect_all()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import io
from pathlib import Path

from PIL import Image


def extract_frames(animated_image_path: Path) -> list[tuple[bytes, int]]:
    with Image.open(animated_image_path) as pil_img:
        frames: list[tuple[bytes, int]] = []
        n_frames: int = getattr(pil_img, "n_frames", 1)

        for i in range(n_frames):
            pil_img.seek(i)
            frame = pil_img.copy()
            with io.BytesIO() as byte_io:
                frame.save(
                    byte_io,
                    format="WEBP",
                    lossless=False,
                    method=0,
                    exact=False,
                    quality=50,
                    alpha_quality=50,
                )
                frame_bytes = byte_io.getvalue()
            duration_ms = int(frame.info.get("duration", 100))
            frames.append((frame_bytes, duration_ms))
            del frame

    return frames



# Copyright (C) 2003-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2005-2006 Dimitur Kirov <dkirov AT gmail.com>
#                         Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2006 Alex Mauer <hawke AT hawkesnest.net>
# Copyright (C) 2006-2007 Travis Shirk <travis AT pobox.com>
# Copyright (C) 2006-2008 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2007 Lukas Petrovicky <lukas AT petrovicky.net>
#                    James Newton <redshodan AT gmail.com>
#                    Julien Pivotto <roidelapluie AT gmail.com>
# Copyright (C) 2007-2008 Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2008 Brendan Taylor <whateley AT gmail.com>
#                    Jonathan Schleifer <js-gajim AT webkeks.org>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import hashlib
import inspect
import json
import logging
import os
import socket
import sys
import unicodedata
import uuid
import weakref
from collections import defaultdict
from collections.abc import Callable
from pathlib import Path
from string import Template

import precis_i18n.codec  # noqa: F401
import qrcode
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Gio
from gi.repository import GLib
from nbxmpp.const import ConnectionProtocol
from nbxmpp.const import ConnectionType
from nbxmpp.errors import StanzaError
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import CommonError
from nbxmpp.structs import ProxyData
from qrcode.image.pil import PilImage as QrcPilImage

from gajim.common import app
from gajim.common import configpaths
from gajim.common import types
from gajim.common.util.standards import get_rfc5646_lang
from gajim.common.util.text import get_random_string

if TYPE_CHECKING:
    from gajim.common.modules.util import LogAdapter

HAS_PYWIN32 = False
if os.name == 'nt':
    try:
        import pywintypes
        import win32con
        import win32file
        HAS_PYWIN32 = True
    except ImportError:
        pass

log = logging.getLogger('gajim.c.helpers')


def sanitize_filename(filename: str) -> str:
    '''
    Sanitize filename of:
     - characters used to obfuscate file names/extensions
     - elements not allowed on Windows
       https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file
    Limit filename length to 50 chars on all systems
    '''

    # Remove right-to-left override U+202E (commonly used to spoof extensions)
    filename = filename.replace("\u202E", "")

    if sys.platform == 'win32':
        blacklist = ['\\', '/', ':', '*', '?', '？', '"', '<', '>', '|', '\0']
        reserved_filenames = [
            'CON', 'PRN', 'AUX', 'NUL', 'COM1', 'COM2', 'COM3', 'COM4', 'COM5',
            'COM6', 'COM7', 'COM8', 'COM9', 'LPT1', 'LPT2', 'LPT3', 'LPT4',
            'LPT5', 'LPT6', 'LPT7', 'LPT8', 'LPT9',
        ]
        filename = ''.join(char for char in filename if char not in blacklist)

        filename = ''.join(char for char in filename if ord(char) > 31)

        filename = unicodedata.normalize('NFKD', filename)
        filename = filename.rstrip('. ')
        filename = filename.strip()

        if all(char == '.' for char in filename):
            filename = f'__{filename}'
        if filename.upper() in reserved_filenames:
            filename = f'__{filename}'
        if len(filename) == 0:
            filename = '__'

    extension = Path(filename).suffix[:10]
    filename = Path(filename).stem
    final_length = 50 - len(extension)

    # Many Filesystems have a limit on filename length: keep it short
    filename = filename[:final_length]

    return f'{filename}{extension}'


def make_path_from_jid(base_path: Path, jid: JID) -> Path:
    assert jid.domain is not None
    domain = jid.domain[:50]

    if jid.localpart is None:
        return base_path / domain

    path = base_path / domain / sanitize_filename(jid.localpart[:50])
    if jid.resource is not None:
        return path / sanitize_filename(jid.resource[:30])
    return path


def generate_qr_code(content: str) -> Gdk.Texture:
    qr = qrcode.QRCode(version=None,
                       error_correction=qrcode.constants.ERROR_CORRECT_L,
                       box_size=6,
                       border=4)
    qr.add_data(content)
    qr.make(fit=True)

    img = qr.make_image(image_factory=QrcPilImage).convert('RGB')
    pixbuf = GdkPixbuf.Pixbuf.new_from_bytes(
        GLib.Bytes.new(img.tobytes()),
        GdkPixbuf.Colorspace.RGB, False, 8,
        img.width, img.height, img.width*3)
    return Gdk.Texture.new_for_pixbuf(pixbuf)


def play_sound(sound_event: str,
               account: str | None = None,
               force: bool = False,
               loop: bool = False) -> None:

    if sound_event is None:
        return
    if (force or account is None or
            allow_sound_notification(account, sound_event)):
        play_sound_file(
            app.settings.get_soundevent_settings(sound_event)['path'], loop)


def check_soundfile_path(file_: str,
                         dirs: list[Path] | None = None
                         ) -> Path | None:
    '''
    Check if the sound file exists

    :param file_: the file to check, absolute or relative to 'dirs' path
    :param dirs: list of knows paths to fallback if the file doesn't exists
                                     (eg: ~/.gajim/sounds/, DATADIR/sounds...).
    :return      the path to file or None if it doesn't exists.
    '''
    if not file_:
        return None
    if Path(file_).exists():
        return Path(file_)

    if dirs is None:
        dirs = [configpaths.get('MY_DATA'),
                configpaths.get('DATA')]

    for dir_ in dirs:
        dir_ = dir_ / 'sounds' / file_
        if dir_.exists():
            return dir_
    return None


def play_sound_file(str_path_to_soundfile: str, loop: bool = False) -> None:
    path_to_soundfile = check_soundfile_path(str_path_to_soundfile)
    if path_to_soundfile is None:
        return

    from gajim.common import sound
    sound.play(path_to_soundfile, loop)


def get_auth_sha(sid: str, initiator: str, target: str) -> str:
    '''
    Return sha of sid + initiator + target used for proxy auth
    '''
    return hashlib.sha1(
        (f'{sid}{initiator}{target}').encode()).hexdigest()


def allow_showing_notification(account: str) -> bool:
    if not app.settings.get('show_notifications'):
        return False
    if app.settings.get('show_notifications_away'):
        return True
    client = app.get_client(account)
    return client.status == 'online'


def allow_sound_notification(account: str, sound_event: str) -> bool:
    if not app.settings.get('sounds_on'):
        return False
    client = app.get_client(account)
    if client.status != 'online' and not app.settings.get('sounddnd'):
        return False
    if app.settings.get_soundevent_settings(sound_event)['enabled']:  # noqa: SIM103
        return True
    return False


def get_optional_features(account: str) -> list[str]:
    features: list[str] = []

    if app.settings.get_account_setting(account, 'request_user_data'):
        features.append(Namespace.TUNE + '+notify')
        features.append(Namespace.LOCATION + '+notify')

    features.append(Namespace.NICK + '+notify')

    client = app.get_client(account)

    if client.get_module('Bookmarks').nativ_bookmarks_used:
        features.append(Namespace.BOOKMARKS_1 + '+notify')
    elif client.get_module('Bookmarks').pep_bookmarks_used:
        features.append(Namespace.BOOKMARKS + '+notify')
    if app.is_installed('AV'):
        features.append(Namespace.JINGLE_RTP)
        features.append(Namespace.JINGLE_RTP_AUDIO)
        features.append(Namespace.JINGLE_RTP_VIDEO)
        features.append(Namespace.JINGLE_ICE_UDP)

    # Give plugins the possibility to add their features
    app.plugin_manager.extension_point('update_caps', account, features)
    return features


def get_global_proxy() -> ProxyData | None:
    proxy_name = app.settings.get('global_proxy')
    if not proxy_name:
        return None
    return get_proxy(proxy_name)


def get_account_proxy(account: str, fallback: bool = True) -> ProxyData | None:
    proxy_name = app.settings.get_account_setting(account, 'proxy')
    if proxy_name:
        return get_proxy(proxy_name)

    if fallback:
        return get_global_proxy()
    return None


def get_proxy(proxy_name: str) -> ProxyData | None:
    if proxy_name == 'no-proxy':
        return ProxyData(type='direct',
                         host='',
                         username=None,
                         password=None)

    try:
        settings = app.settings.get_proxy_settings(proxy_name)
    except ValueError:
        return None

    username, password = None, None
    if settings['useauth']:
        username, password = settings['user'], settings['pass']

    return ProxyData(type=settings['type'],
                     host=f"{settings['host']}:{settings['port']}",
                     username=username,
                     password=password)


def determine_proxy(account: str | None = None) -> ProxyData | None:

    if account is not None:
        return get_account_proxy(account)

    global_proxy = get_global_proxy()
    if global_proxy is not None:
        return global_proxy

    # When there is no global proxy and at least one active account does
    # not use a proxy, we assume no proxy is necessary.

    proxies: list[ProxyData] = []
    for client in app.get_clients():
        account_proxy = get_account_proxy(client.account, fallback=False)
        if account_proxy is None:
            return None

        proxies.append(account_proxy)

    return proxies[0] if proxies else None


def load_json(path: Path,
              key: str | None = None,
              default: Any | None = None) -> Any:
    try:
        with path.open('r', encoding='utf8') as file:
            json_dict = json.loads(file.read())
    except Exception:
        log.exception('Parsing error')
        return default

    if key is None:
        return json_dict
    return json_dict.get(key, default)


def dump_json(path: Path, data: dict[Any, Any]) -> None:
    """Save a JSON-serializable object to a .json file."""
    try:
        with path.open('w', encoding='utf8') as file:
            json.dump(data, file)
    except Exception:
        log.exception('Error while trying to dump JSON')


def file_is_locked(path_to_file: str) -> bool:
    '''
    Return True if file is locked
    NOTE: Windows only.
    '''
    if os.name != 'nt':
        return False

    if not HAS_PYWIN32:
        return False

    secur_att = pywintypes.SECURITY_ATTRIBUTES()
    secur_att.Initialize()

    try:
        # try create a handle for READING the file
        hfile = win32file.CreateFile(
            path_to_file,
            win32con.GENERIC_READ,  # open for reading
            0,  # do not share with other proc
            secur_att,
            win32con.OPEN_EXISTING,  # existing file only
            win32con.FILE_ATTRIBUTE_NORMAL,  # normal file
            0)  # no attr. template
    except pywintypes.error:
        return True
    else:  # in case all went ok, close file handle
        hfile.Close()
        return False


def get_resource(account: str) -> str | None:
    resource = app.settings.get_account_setting(account, 'resource')
    if not resource:
        return None

    resource = Template(resource).safe_substitute(
        {'hostname': socket.gethostname(),
         'rand': get_random_string()})
    app.settings.set_account_setting(account, 'resource', resource)
    return resource


def to_user_string(error: CommonError | StanzaError) -> str:
    text = error.get_text(get_rfc5646_lang())
    if text:
        return text

    condition = error.condition
    if error.app_condition is not None:
        return f'{condition} ({error.app_condition})'
    return condition


class Observable:
    def __init__(self, log_: logging.Logger | LogAdapter | None = None):
        self._log = log_
        self._callbacks: types.ObservableCbDict = defaultdict(list)

    def __disconnect(self,
                     obj: Any,
                     signals: set[str] | None = None
                     ) -> None:

        def _remove(handlers: list[weakref.WeakMethod[types.AnyCallableT]]
                    ) -> None:

            for handler in list(handlers):
                func = handler()
                # Don’t remove dead weakrefs from the handler list
                # notify() will remove dead refs, and __disconnect()
                # can be called from inside notify(), this can lead
                # to race conditions where later notfiy tries to remove
                # a dead ref which is not anymore in the list.
                if func is not None and func.__self__ is obj:
                    handlers.remove(handler)

        if signals is None:
            for handlers in self._callbacks.values():
                _remove(handlers)

        else:
            for signal in signals:
                _remove(self._callbacks.get(signal, []))

    def disconnect_signals(self) -> None:
        self._callbacks = defaultdict(list)

    def multi_disconnect(self,
                         obj: Any,
                         signals: set[str] | None
                         ) -> None:

        self.__disconnect(obj, signals)

    def disconnect_all_from_obj(self, obj: Any) -> None:
        self.__disconnect(obj)

    def disconnect(self, obj: Any) -> None:
        self.disconnect_all_from_obj(obj)

    def disconnect_signal(self, obj: Any, signal: str) -> None:
        self.__disconnect(obj, {signal})

    def connect_signal(self,
                       signal_name: str,
                       func: types.AnyCallableT) -> None:
        if not inspect.ismethod(func):
            raise ValueError('Only bound methods allowed')

        weak_func = weakref.WeakMethod(func)

        if weak_func in self._callbacks[signal_name]:
            # Don’t register handler multiple times
            return

        self._callbacks[signal_name].append(weak_func)

    def connect(self,
                signal_name: str,
                func: types.AnyCallableT) -> None:
        self.connect_signal(signal_name, func)

    def multi_connect(self, signal_dict: dict[str, types.AnyCallableT]):
        for signal_name, func in signal_dict.items():
            self.connect_signal(signal_name, func)

    def notify(self, signal_name: str, *args: Any, **kwargs: Any):
        signal_callbacks = self._callbacks.get(signal_name)
        if not signal_callbacks:
            return

        if self._log is not None:
            self._log.info('Signal: %s', signal_name)

        for weak_method in list(signal_callbacks):
            func = weak_method()
            if func is None:
                self._callbacks[signal_name].remove(weak_method)
                continue
            func(self, signal_name, *args, **kwargs)


def write_file_async(
        path: Path,
        data: bytes,
        callback: Callable[[bool, GLib.Error | None, Any], Any],
        user_data: Any | None = None):

    def _on_write_finished(outputstream: Gio.OutputStream,
                           result: Gio.AsyncResult,
                           _data: bytes) -> None:
        try:
            successful, _bytes_written = outputstream.write_all_finish(result)
        except GLib.Error as error:
            callback(False, error, user_data)
        else:
            callback(successful, None, user_data)

    def _on_file_created(file: Gio.File, result: Gio.AsyncResult) -> None:
        try:
            outputstream = file.create_finish(result)
        except GLib.Error as error:
            callback(False, error, user_data)
            return

        # Pass data as user_data to the callback, because
        # write_all_async() takes no reference to the data
        # and python gc collects it before the data is written
        outputstream.write_all_async(data,
                                     GLib.PRIORITY_DEFAULT,
                                     None,
                                     _on_write_finished,
                                     data)

    file = Gio.File.new_for_path(str(path))
    file.create_async(Gio.FileCreateFlags.PRIVATE,
                      GLib.PRIORITY_DEFAULT,
                      None,
                      _on_file_created)

def load_file_async(path: Path,
                    callback: Callable[[bytes | None,
                                        GLib.Error | None,
                                        Any], Any],
                    user_data: Any | None = None) -> None:

    def _on_load_finished(file: Gio.File,
                          result: Gio.AsyncResult) -> None:

        try:
            _, contents, _ = file.load_contents_finish(result)
        except GLib.Error as error:
            callback(None, error, user_data)
        else:
            # "contents" may be an empty bytes object
            callback(contents or None, None, user_data)

    file = Gio.File.new_for_path(str(path))
    file.load_contents_async(None, _on_load_finished)


def get_x509_cert_from_gio_cert(cert: Gio.TlsCertificate) -> x509.Certificate:
    glib_bytes = GLib.ByteArray.free_to_bytes(cert.props.certificate)
    return x509.load_der_x509_certificate(
        glib_bytes.get_data(), default_backend())


def get_custom_host(
    account: str
) -> tuple[str, ConnectionProtocol, ConnectionType] | None:

    if not app.settings.get_account_setting(account, 'use_custom_host'):
        return None
    host = app.settings.get_account_setting(account, 'custom_host')
    port = app.settings.get_account_setting(account, 'custom_port')
    type_ = app.settings.get_account_setting(account, 'custom_type')

    if host.startswith(('ws://', 'wss://')):
        protocol = ConnectionProtocol.WEBSOCKET
    else:
        host = f'{host}:{port}'
        protocol = ConnectionProtocol.TCP

    return (host, protocol, ConnectionType(type_))


def warn_about_plain_connection(account: str,
                                connection_types: list[ConnectionType]
                                ) -> bool:
    warn = app.settings.get_account_setting(
        account, 'confirm_unencrypted_connection')
    return any(type_.is_plain and warn for type_ in connection_types)


def get_uuid() -> str:
    return str(uuid.uuid4())


def idle_add_once(func: Callable[..., Any], *args: Any) -> None:
    def wrapper():
        func(*args)
        return False

    GLib.idle_add(wrapper)


def timeout_add_once(ms: int, func: Callable[..., Any], *args: Any) -> None:
    def wrapper():
        func(*args)
        return False

    GLib.timeout_add(ms, wrapper)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Types for typechecking

from typing import Any
from typing import TYPE_CHECKING
from typing import Union

import weakref
from collections.abc import Callable

import nbxmpp
from gi.repository import GdkPixbuf
from nbxmpp.const import PresenceShow
from nbxmpp.protocol import JID
from nbxmpp.structs import BookmarkData

from gajim.common.const import PresenceShowExt

if TYPE_CHECKING:
    from nbxmpp.client import Client as NBXMPPClient  # type: ignore # noqa: F401

    from gajim.common.client import Client
    from gajim.common.modules.contacts import BareContact
    from gajim.common.modules.contacts import CommonContact  # type: ignore # noqa: F401
    from gajim.common.modules.contacts import GroupchatContact
    from gajim.common.modules.contacts import GroupchatParticipant
    from gajim.common.modules.contacts import ResourceContact
    from gajim.common.settings import Settings
    from gajim.plugins.pluginmanager import PluginManager
    from gajim.plugins.repository import PluginRepository

    from gajim.gtk.css_config import CSSConfig

    ContactT = (BareContact |
                ResourceContact |
                GroupchatContact |
                GroupchatParticipant)

PluginManagerT = Union['PluginManager']
PluginRepositoryT = Union['PluginRepository']

ConnectionT = Union['Client']
CSSConfigT = Union['CSSConfig']

# PEP
PEPNotifyCallback = Callable[[nbxmpp.JID, nbxmpp.Node], None]
PEPHandlersDict = dict[str, list[PEPNotifyCallback]]

# Plugins
PluginExtensionPoints = dict[str, tuple[Callable[..., None] | None,
                                        Callable[..., None] | None]]

SettingsT = Union['Settings']

BookmarksDict = dict[JID, BookmarkData]

GdkPixbufType = GdkPixbuf.Pixbuf | GdkPixbuf.PixbufAnimation

AnyCallableT = Callable[..., Any]
ObservableCbDict = dict[str, list[weakref.WeakMethod[AnyCallableT]]]

BareContactT = Union['BareContact']
ChatContactT = Union['BareContact', 'GroupchatContact', 'GroupchatParticipant']
OneOnOneContactT = Union['BareContact', 'GroupchatParticipant']
GroupchatContactT = Union['GroupchatContact']

PresenceShowT = PresenceShowExt | PresenceShow



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

from collections import defaultdict

import nbxmpp
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import BookmarkData
from nbxmpp.structs import MessageProperties

from gajim.common import app
from gajim.common import types
from gajim.common.events import MucUserBlockChanged
from gajim.common.modules.base import BaseModule
from gajim.common.modules.util import event_node

GAJIM_BLOCKING_NS = "urn:gajim:blocking:1"

BlockedMucUsersT = dict[JID, set[str]]


class MucBlocking(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)
        self._register_pubsub_handler(self._bookmark_1_event_received)

        self._blocked_muc_users: BlockedMucUsersT = defaultdict(set)
        occupants = app.storage.archive.get_blocked_occupants(self._account)
        for occupant in occupants:
            self._blocked_muc_users[occupant.remote.jid].add(occupant.id)

    def _sync_muc_blocks(self) -> bool:
        if not self._client.get_module("Bookmarks").nativ_bookmarks_used:
            return False

        return app.settings.get_account_setting(self._account, "sync_muc_blocks")

    @event_node(Namespace.BOOKMARKS_1)
    def _bookmark_1_event_received(
        self, _con: types.NBXMPPClient, stanza: Any, properties: MessageProperties
    ) -> None:
        if not self._sync_muc_blocks():
            return

        event = properties.pubsub_event
        assert event is not None
        if event.deleted or event.purged:
            self._log.info("Node deleted or purged")
            self._blocked_muc_users.clear()
            app.storage.archive.set_block_occupant(self._account, None, [], False)
            app.ged.raise_event(MucUserBlockChanged(account=self._account))
            return

        if event.retracted:
            self._log.info("Retracted: %s", event.id)
            assert event.id is not None
            try:
                jid = JID.from_string(event.id)
            except Exception as error:
                self._log.warning(
                    "Unable to parse retracted id: %s %s", event.id, error
                )
                self._log.warning(stanza)
                return

            self._blocked_muc_users.pop(jid, None)
            app.storage.archive.set_block_occupant(self._account, jid, [], False)
            app.ged.raise_event(MucUserBlockChanged(account=self._account, jid=jid))
            return

        self._log.info("Update received: %s", event.id)
        bookmark = cast(BookmarkData, event.data)
        if self._act_on_changed_blocks(bookmark):
            app.ged.raise_event(
                MucUserBlockChanged(account=self._account, jid=bookmark.jid)
            )

    def get_blocking_list(self, jid: JID) -> set[str]:
        return self._blocked_muc_users.get(jid) or set()

    def is_blocked(self, jid: JID, occupant_id: str | None) -> bool:
        if not occupant_id:
            return False

        return occupant_id in self._blocked_muc_users[jid]

    def pass_bookmarks(self, bookmarks: list[BookmarkData]) -> None:
        if not self._sync_muc_blocks():
            return

        self._log.info("Received bookmarks")
        self._blocked_muc_users.clear()
        blocks_changed = False
        for bookmark in bookmarks:
            blocks_changed |= self._act_on_changed_blocks(bookmark)

        if blocks_changed:
            app.ged.raise_event(MucUserBlockChanged(account=self._account))

    def _act_on_changed_blocks(self, bookmark: BookmarkData) -> bool:
        try:
            new_occupant_ids = parse_blocked_occupants(bookmark.extensions)
        except Exception as error:
            self._log.warning("Unable to parse blocking extension:%s", error)
            self._log.warning(bookmark)
            return False

        if new_occupant_ids is None:
            # None means there was never blocking information published
            # to this bookmark, this is different from an empty blocking list
            return False

        jid = bookmark.jid
        old_occupant_ids = self._blocked_muc_users[jid]
        if new_occupant_ids == old_occupant_ids:
            self._log.info("%s: no changes found", jid)
            return False

        self._blocked_muc_users[jid] = new_occupant_ids

        added = new_occupant_ids - old_occupant_ids
        removed = old_occupant_ids - new_occupant_ids

        if added:
            app.storage.archive.set_block_occupant(
                self._account, jid, list(added), True
            )
        if removed:
            app.storage.archive.set_block_occupant(
                self._account, jid, list(removed), False
            )

        self._log.info(
            "Received blocks for %s, added: %s, removed: %s", jid, added, removed
        )
        return True

    def set_block_occupants(
        self, jid: JID, occupant_ids: list[str], block: bool
    ) -> None:
        if not occupant_ids:
            self._log.warning(
                "Unable to block contact because occupant id is not available"
            )
            return

        for occupant_id in occupant_ids:
            self._log.info(
                "%s user: %s %s", "Block" if block else "Unblock", jid, occupant_id
            )

        bookmark = self._client.get_module("Bookmarks").get_bookmark(jid)
        if bookmark is None:
            self._log.warning("Unable to block user, no bookmark found")
            return

        current_blocks = self._blocked_muc_users[jid]
        if block:
            current_blocks.update(occupant_ids)
        else:
            current_blocks.difference_update(occupant_ids)

        app.storage.archive.set_block_occupant(self._account, jid, occupant_ids, block)

        app.ged.raise_event(MucUserBlockChanged(account=self._account, jid=jid))

        if not self._sync_muc_blocks():
            return

        extensions = create_extensions_node(bookmark.extensions, current_blocks)
        self._client.get_module("Bookmarks").modify(jid, extensions=extensions)

    def merge_blocks(self) -> None:
        if not self._client.get_module("Bookmarks").nativ_bookmarks_used:
            return

        for bookmark in self._client.get_module("Bookmarks").bookmarks:
            local_occupant_ids = self.get_blocking_list(bookmark.jid)

            try:
                remote_occupant_ids = (
                    parse_blocked_occupants(bookmark.extensions) or set()
                )
            except Exception as error:
                self._log.warning("Unable to parse blocking extension:%s", error)
                self._log.warning(bookmark)
                continue

            remote_occupant_ids.update(local_occupant_ids)
            if not remote_occupant_ids:
                # Nothing to sync
                continue

            extensions = create_extensions_node(
                bookmark.extensions, remote_occupant_ids
            )
            self._client.get_module("Bookmarks").modify(
                bookmark.jid, extensions=extensions
            )

    def cleanup(self) -> None:
        BaseModule.cleanup(self)
        self._blocked_muc_users.clear()


def create_extensions_node(
    extensions: nbxmpp.Node | None, blocked_occupants: set[str]
) -> nbxmpp.Node:
    if extensions is None:
        extensions = nbxmpp.Node("extensions", attrs={"xmlns": "urn:xmpp:bookmarks:1"})
    else:
        extensions = nbxmpp.Node(
            node=str(extensions), attrs={"xmlns": "urn:xmpp:bookmarks:1"}
        )

    nodes = list(extensions.iterTags("blocking", namespace=GAJIM_BLOCKING_NS))
    for node in nodes:
        extensions.delChild(node)

    blocking = build_blocking_node(blocked_occupants)
    extensions.addChild(node=blocking)
    return extensions


def build_blocking_node(occupant_ids: set[str]) -> nbxmpp.Node:
    blocking = nbxmpp.Node(tag="blocking", attrs={"xmlns": GAJIM_BLOCKING_NS})
    for occupant_id in occupant_ids:
        blocking.addChild("occupant", attrs={"id": occupant_id})
    return blocking


def parse_blocked_occupants(extension: nbxmpp.Node | None) -> set[str] | None:
    if extension is None:
        return None

    blocking = extension.getTag("blocking", namespace=GAJIM_BLOCKING_NS)
    if blocking is None:
        return None

    occupant_ids: set[str] = set()
    for occupant in blocking.getTags("occupant"):
        occupant_id = occupant.getAttr("id")
        if not occupant_id:
            continue
        occupant_ids.add(occupant_id)

    return occupant_ids
